// Add repeating CW message to PCM audio streaming through

#define _GNU_SOURCE 1
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <tgmath.h>
#include <string.h>
#include <complex.h>
#include <limits.h>
#include "dspfec.h"

// set defaults here
double Sample_rate = 48000;
double Speed = 15; // wpm
double Frequency = 470.; // fits lower notch in BPSK spectrum
double Amplitude = .1; // -20 dB from max; don't overdo it
double Risetime;

int Element; // Element duration in samples
int Clip; // Number of output sample clips
int Risetime_samples;
complex double Tone_oscillator;
complex double Shaper;
complex double Tone_step;
complex double Shaper_step;

void sincos(double,double *,double *);

int Generate; // When set, generate audio; don't act as a filter

struct option Options[] = {
  { "sample-rate", 1, NULL, 'r' },

  { "generate", 0, NULL, 'g' },
  { "no-filter", 0, NULL, 'g' },

  { "frequency", 1, NULL, 'c' },
  { "carrier",1,NULL,'c' },
  { "pitch",1,NULL,'c' },

  { "speed",1,NULL,'s' },
  { "wpm",1,NULL,'s' },

  { "amplitude",1,NULL,'a' },
  { "level",1,NULL,'a' },

  { "rise-time",1,NULL,'R' },
  { "risetime",1,NULL,'R' },

  { "verbose",0,NULL,'v' },
  { 0, 0, NULL, 0}
};


char *Morse_letters[] = {
  "._","_...","_._.","_..",".",".._.","__.",
  "....","..",".___","_._","._..","__","_.","___",".__.","__._",
  "._.","...","_",".._","..._",".__","_.._","_.__","__.."
};

char *Morse_numbers[] = {
  "_____",".____","..___","...__","...._",".....","_....","__...","___..","____."
};


void send_cw(char *);
void dot();
void dash();
void wordspace();
void letterspace();
double *Envelope;
double kaiser(int n,int m,double beta);


int main(int argc,char *argv[]){
  int c,i;
  int option_index;
  double ang;

  // Set defaults
  Sample_rate = 48000;
  Speed = 15; // wpm
  Frequency = 470.; // fits lower notch in BPSK spectrum
  Amplitude = .1; // -20 dB from max; don't overdo it
  Risetime = .005; // 5 ms recommended by ARRL
  while((c = getopt_long(argc,argv,"R:vf:c:s:a:r:g",Options,&option_index)) != EOF){
    switch(c){
    case 'r':
      Sample_rate = atoi(optarg);
      break;
    case 'g': // Generate; don't act as a filter
      Generate++;
      break;
    case 'f':
    case 'c':
      Frequency = atof(optarg);
      break;
    case 's':
      Speed = atof(optarg);
      break;
    case 'a':
      Amplitude = pow(10.,-fabs(atof(optarg))/20.);
      break;
    case 'R':
      Risetime = atof(optarg);
      break;
    case 'v':
      Verbose++;
      break;
    default:
      fprintf(stderr,"Usage: %s: [-g|--generate] [-v|--verbose] | [-R|--rise-time <time>] [-c|--carrier <carrier_freq>] [-s|--speed <wpm>] [-a <amplitude_dB>] [-r|--sample-rate <rate>] message ...\n",argv[0]);
      exit(1);
    }
  }
  cpu_id();
  Element = Sample_rate * 1.2/Speed;
  Risetime_samples = Risetime * Sample_rate;

  Tone_oscillator = 1; // 0 deg
  Shaper = -1; // 180 degrees = Off; +1 is full-on
  ang = 2*M_PI*Frequency/Sample_rate;
  Tone_step = cos(ang) + I*sin(ang);
  ang = M_PI / Risetime_samples;
  Shaper_step = cos(ang) + I*sin(ang);


  if(Verbose){
    fprintf(stderr,"Pitch %.1lf Hz; Speed %.1lf wpm;",Frequency,Speed);
    fprintf(stderr," element %.4lf sec; risetime = %.4lf sec\n",1.2/Speed,Risetime);
    fprintf(stderr,"Tone increment = %lf deg/sample; shaper increment %lf deg/sample\n",
	    (180./M_PI) * carg(Tone_step),
	    (180./M_PI) * carg(Shaper_step));
  }
  if(argc <= optind){
    fprintf(stderr,"%s: message missing\n",argv[0]);
    exit(1);
  }

  while(1){
    for(i=optind;i<argc;i++){
      send_cw(argv[i]);
      send_cw(" "); // delay before repeat; also protect against null message
    }
  }
  exit(0);

}


void send_cw(char *message){
  int c;
  char *code,*cp;

  if(message == NULL)
    return;

  while((c = *message++) != '\0'){
    if(islower(c))
      c = toupper(c);
    if(isupper(c)){
	c -= 'A';
	if(c >= 26)
	  continue;
	code = Morse_letters[c];
    } else if(isdigit(c)){
      c -= '0';
      if(c > 9)
	continue;
      code = Morse_numbers[c];
    } else if(isspace(c)){
      code = " ";
    } else {
      switch(c){
      case '.':
	code = "._._._";
	break;
      case ',':
	code = "__..__";
	break;
      case '?':
	code = "..__..";
	break;
      case '\'':
	code = ".____.";
	break;
      case '!':
	code = "_._.__";
	break;
      case '/':
	code = "_.._.";
	break;
      case '(':
	code = "_.__."; // prosign KN - go ahead specified station only
	break;
      case ')':
	code = "_.__._";
	break;
      case '&':
	code = "._..."; // prosign AS - wait
	break;
      case ':':
	code = "___...";
	break;
      case ';':
	code = "_._._.";
	break;
      case '=':
	code = "_..._"; // prosign BT - filler
	break;
      case '+':
	code = "._._."; // prosign AR - end of message
	break;
      case '-':
	code = "_...._";
	break;
      case '_':
	code = "..__._";
	break;
      case '"':
	code = "._.._.";
	break;
      case '$':
	code = "..._.._";
	break;
      case '@':
	code = ".__._."; // prosign AC - commercial at-sign
	break;
      case '\\':
	code = "........"; // error signal
	break;
      default:
	continue;
      }
    }

    cp = code;
    while((c = *cp++) != '\0'){
      switch(c){
      case '.':
	dot();
	break;
      case '_':
      case '-':
	dash();
	break;
      case ' ':
	wordspace();
	break;
      }
    }
    letterspace();
  }
}

// tone on -- generate tone, optionally add input, write
void tone_on(int nsamp){
  int i;
  signed short *samples;
  signed short tone[nsamp];
  int sample_sum; // 32 bits
  int clip;
  
  for(i=0;i<Risetime_samples;i++){
    // Shaper rotates from -PI to +PI (cos = -1 to +1)
    tone[i] = SHRT_MAX * 0.5 * (creal(Shaper) + 1) * Amplitude * creal(Tone_oscillator);
    Tone_oscillator *= Tone_step;
    Shaper *= Shaper_step;
  }
  for(;i<nsamp-Risetime_samples;i++){
    tone[i] = SHRT_MAX * Amplitude * creal(Tone_oscillator);
    Tone_oscillator *= Tone_step;
  }
  for(;i<nsamp;i++){
    // Shaper rotates from +PI back to -PI (cos = +1 to -1)
    tone[i] = SHRT_MAX * 0.5 * (creal(Shaper) + 1) * Amplitude * creal(Tone_oscillator);
    Tone_oscillator *= Tone_step;
    Shaper *= Shaper_step;
  }

  if(Generate){
    fwrite(tone,sizeof(signed short),nsamp,stdout);
    return;
  }
  // Add tone to stdin and clip
  samples = alloca(nsamp*sizeof(signed short));
  fread(samples,sizeof(signed short),nsamp,stdin);    
  if(feof(stdin))
    exit(0);
  clip = 0;
  for(i=0;i<nsamp;i++){
    sample_sum = samples[i] + tone[i];
    if(sample_sum > SHRT_MAX){
      sample_sum = SHRT_MAX;
      clip++;
    } else if (sample_sum < -SHRT_MAX){ // Disallow -32768, can cause troubles
      sample_sum = -SHRT_MAX;
      clip++;
    }
    samples[i] = sample_sum;
  }
  if(clip)
    fprintf(stderr,"addcw: Warning - audio clipping is occurring; reduce input or CW level\n");
  
  fwrite(samples,sizeof(signed short),nsamp,stdout);
}

// Tone off - copy samples without change
void tone_off(int nsamp){
  signed short samples[nsamp];
  int i;

  // Keep tone phase coherent - because I can!
  for(i=0;i<nsamp;i++)
    Tone_oscillator *= Tone_step;

  if(Generate){
    memset(samples,0,sizeof(samples));
  } else {
    fread(samples,sizeof(signed short),nsamp,stdin);
    if(feof(stdin))
      exit(0);
  }
  fwrite(samples,sizeof(signed short),nsamp,stdout);
}

// Carrier on for one element
void dot(){
  tone_on(Element);
  tone_off(Element);
}

// Carrier on for three elements
void dash(){
  tone_on(3*Element);
  tone_off(Element);
}

void letterspace(){
  tone_off(2*Element);
}

void wordspace(){
  tone_off(6*Element);
}

