/* Add gaussian noise to a PCM audio signal
 * Jan 2002, Phil Karn, KA9Q
 * Updated April 2009, Phil Karn, KA9Q
 * Copyright 2009, Phil Karn. This software may be used under the terms
 * of the GNU Public License (GPL).
 */

#include <stdio.h>
#include <tgmath.h>
#include <limits.h>
#include <memory.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include "dspfec.h"

struct option Options[] = {
  { "blocksize", 1, NULL, 'B' },
  { "bit-rate", 1, NULL, 'b' },
  { "fade-period",1, NULL, 'f' },
  { "sample-rate", 1, NULL, 'r' },
  { "ebn0", 1, NULL, 'e' },
  { "snr", 1, NULL, 'n' },
  { "verbose", 0, NULL, 'v' },
  { 0, 0, NULL, 0}
};

/* Default settings */
int Blocksize = 48000;

int Bitrate = 500;
int Samprate = 48000;
double Ebn0 = 0; // No default
double Fade_period = 0;

int main(int argc,char *argv[]){
  signed short *samples,sample;
  int cnt,i;
  double s,signal_ampl,noise_ampl,gain;
  long long signal_energy;
  int option_index = 0;
  double snr = 0;
  long t;
  double fadephase = 0;
  
  while((i = getopt_long(argc,argv,"B:f:b:r:e:n:v",Options,&option_index)) != EOF){
    switch(i){
    case 'B':
      Blocksize = atoi(optarg);
      break;
    case 'f':
      Fade_period = atof(optarg);
      break;
    case 'b':
      Bitrate = atof(optarg);
      break;
    case 'r':
      Samprate = atof(optarg);
      break;
    case 'e':
      Ebn0 = pow(10.,atof(optarg)/10.); /* As energy ratio */
      break;
    case 'n':
      snr = pow(10.,atof(optarg)/20.); /* As amplitude ratio */
      break;
    case 'v':
      Verbose++;
      break;
    }
  }
  if(Blocksize == 0){
    fprintf(stderr,"%s: Blocksize 0 invalid; using default of 1 second\n",argv[0]);
    Blocksize = Samprate;
  }
  if((samples = malloc(Blocksize * sizeof(signed short))) == NULL){
    fprintf(stderr,"%s: Can't malloc buffer of %lu bytes\n",argv[0],Blocksize*sizeof(signed short));
    exit(1);
  }

  if(snr == 0){
    /* Compute from bitrate, sample rate and Ebn0 */
    if(Ebn0 == 0 || Bitrate == 0 || Samprate == 0){
      fprintf(stderr,"%s: Must specify --snr or --bit-rate, --sample-rate and --ebn0\n",argv[0]);
      exit(1);
    }
    snr = sqrt(2 * Ebn0 * Bitrate / Samprate);
  } else {
    Ebn0 = Samprate*snr*snr/(2*Bitrate);
  }
  if(Verbose){
    fprintf(stderr,"%s: sample rate %d; bit rate %d; Eb/No = %.1f dB; SNR = (Eb/No - %.1f dB) = %.1f dB; fade rate %.1f sec\n",
	    argv[0],Samprate,Bitrate,10*log10(Ebn0),10*log10(2*Samprate/Bitrate),20*log10(snr),Fade_period);
  }

  time(&t);
  zigset(t);
  if(Fade_period != 0)
    Fade_period = (2*M_PI/Fade_period)/Samprate; /* Convert from cycles/sec to radians/sample */

  while((cnt = fread(samples,sizeof(unsigned short),Blocksize,stdin)) > 0){
    signal_energy = sumsq(samples,cnt);

    /* Compute signal, noise amplitudes */
    signal_ampl = sqrt((double)signal_energy/cnt);
    noise_ampl = signal_ampl/snr;

    /* Set gain to make signal + noise amplitude 12 dB below clipping */
    gain = SHRT_MAX / (4 * (signal_ampl + noise_ampl));
    if(Verbose > 1)
      fprintf(stderr,"%s: signal %.0f RMS, noise %.0f RMS, gain %f\n",argv[0],signal_ampl,noise_ampl,gain);

    if(Fade_period == 0){
      for(i=0;i<cnt;i++){ /* No fading */
	s = gain * (noise_ampl * rnor() + (double)samples[i]);
	sample = (s > 32767) ? 32767 : (s < -32768 ? -32768 : s);
	fwrite(&sample,sizeof(sample),1,stdout);
      }
    } else {
      for(i=0;i<cnt;i++){ /* fading */
	double fadegain;
	fadegain = M_SQRT2 * cos(fadephase);
	fadephase += Fade_period;
	s = gain * (noise_ampl * rnor() + fadegain * (double)samples[i]);
	sample = (s > 32767) ? 32767 : (s < -32768 ? -32768 : s);
	fwrite(&sample,sizeof(sample),1,stdout);
      }
    }
  }
  exit(0);
}
