#define HDLC_FLAG 0x7e
#define HDLC_ESCAPE 0x7d
#define CRCPOLY 0x04c11db7
#define FRAME_SIZE 1024

static long update_crc(long,char);

// Asynchronous HDLC stream decoder
// November 2010, Phil Karn, KA9Q
// Call async_hdlc_decode() with each incoming character and check the return code:
//   0 -> frame decode still in progress; nothing to do
//  -1 -> error; nothing to do
// = 4 -> A good frame has been decoded and placed in HDLC_Frame[]; go read it. The return value is the length in bytes, including the 4-byte CRC.

// This decoder handles only one stream at a time. If you need it to be re-entrant, the static state variables inside the function
// should be moved into a state structure that is handed as a pointer to each call.

char HDLC_Frame[FRAME_SIZE];

int async_hdlc_decode(char c){
  static int escape = 0;
  static long crc = 0xffffffff;
  static int write_pointer = 0;
  static int search_mode = 1; // when true, ignore everything up to the next flag
  int length;

  switch(c){
  case HDLC_ESCAPE:
    if(search_mode)
      return 0;
    escape = 1;
    length = 0;
    break;
  case HDLC_FLAG:
    if(write_pointer == 0){
      length = 0;       // Back-to-back flags
    } else if(write_pointer < 4){
      length = -1;      // Runt
    } else if((crc & 0xffffffff) != 0){
      length = -1;      // CRC failure
    } else {
      length = write_pointer; // Good frame
    }
    write_pointer = 0; // Reset frame assembly
    crc = 0xffffffff; // Reset CRC
    escape = 0;
    search_mode = 0;
    break;
  default:
    if(search_mode)
      return 0;
    if(escape){
      c ^= 0x20; // Flip bit 5
      escape = 0;
    }
    if(write_pointer >= FRAME_SIZE){
      // Frame is too big; flush it and search for the next flag
      search_mode = 1;
      length = -1;
    } else {
      HDLC_Frame[write_pointer++] = c;
      crc = update_crc(crc,c);
      length = 0;
    }
    break;
  }
  return length;
}
static long update_crc(long crc,char c){
  int i;
  int feedback;

  for(i=0;i<8;i++){
    // if crc.31 xor c.0 is 1, feedback = 0xffffffff, else 0
    feedback = -( ((crc >> 31) ^ c) & 1);
    c >>= 1;
    crc = (crc << 1) ^ (CRCPOLY & feedback);
  }
  return crc;
}

// Simple test program. Turn off this #if if you don't need it.
#if 1
#include <stdio.h>
#include <stdlib.h>

int main(){
  int c,i;
  int length;

  while((c = getchar()) != EOF){
    length = async_hdlc_decode(c);
    if(length >0){
      printf("Good frame, %d bytes:\n",length);
      for(i=0;i<length;i++){
	printf(" %02x",HDLC_Frame[i] & 0xff);
      }
      printf("\n");
    } else if(length == -1){
      printf("Error!\n");
    }
  }
  exit(0);
}

#endif
