// Convolutional de-interleave, Viterbi decode
// June 2009, Phil Karn, KA9Q
// Copyright 2009, Phil Karn. This software may be used under the terms
// of the GNU Public License (GPL).

// The decoder feeds incoming symbols to all de-interleaver phases
// until one decodes. Then it declares lock and continues feeding symbols
// to the successful phase until an HDLC error occurs, then it drops back to trying all of them.

// The CPU cycles required to do all phases is fairly small at this low data rate,
// so the only real reason to not do brute force was to reduce the risk of noise
// generating a spurious frame with valid CRC. Now that a 32-bit CRC is used,
// the chances of this are pretty small.

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <memory.h>
#include <limits.h>
#include <pthread.h>
#include "dspfec.h"
#include "dpsk.h"

#define DECODER_TEST 0
#define MAX_HDLC_SIZE 600
#define DECODER_SIZE 32768 // This MUST be a power of two, and bigger than one interleaver span (N1*N1)
#define DECODER_MAGIC 0x4587FE17

#if DECODER_SIZE <= (N1*N1)
#error "DECODER_SIZE must be larger than one interleaver span, N1*N1"
#endif

#if (DECODER_SIZE & (DECODER_SIZE-1)) != 0
#error "DECODER_SIZE must be a power of two"
#endif

enum hdlc_state { OK=0,CRC_FAIL,ABORT,RUNT,TOO_BIG };

// There's one of these for each possible phase in the de-interleaver
struct decoder {
  unsigned int magic;
  enum hdlc_state state;
  unsigned char history; // History of last 8 bits received, newest bit on right
  unsigned int outbits; // Number of bits in current frame
  unsigned int crc;     // Must be 32 bits
  unsigned char buffer[MAX_HDLC_SIZE]; // Current frame
  unsigned long long goodframes;
  void *vit; // Viterbi decoder
};

struct decoder Decoders[N1];
extern pthread_mutex_t Stderr_mutex;
unsigned long long Decoder_symbols_processed = 0;
unsigned long long Decoded_syms_processed = 0;
unsigned long long Goodbytes = 0;
unsigned long long Goodframes = 0;
int Locked = -1;
extern int All_phases;
unsigned int Histogram_interval;
unsigned int Framing_mode = 0;
long long Histogram[256];
int Runts, Aborts, CRC_fails,Too_big;
extern int Hdlc, Clean;

static inline unsigned int permute(unsigned int x){
  return bitreversen(x,N1);
}

unsigned int process_hdlc_bit(int phase,unsigned int);
void decoder_stats(void);


int Decoder_Q[DECODER_SIZE];
unsigned int Decoder_write = 0;
unsigned int Decoder_read = 0;
extern int EOD_encountered;

unsigned int Unprocessed_symbols;
unsigned int Highwater;
pthread_mutex_t Sym_mutex;
pthread_cond_t Sym_q_not_empty;
pthread_cond_t Sym_q_not_full;

// Decoder thread
void *decoder(void *p){
  int a,b,c,interleaver_phase;
  unsigned char symbol_buffer[N1],output[16];
  int queue_length,i;

  interleaver_phase = 0;
  while(1){
    // Block if the queue is empty
    pthread_mutex_lock(&Sym_mutex);
    while(Unprocessed_symbols == 0){
      if(EOD_encountered){
	pthread_mutex_unlock(&Sym_mutex);
	goto eof;
      }
      pthread_cond_wait(&Sym_q_not_empty,&Sym_mutex);
    }
    // Cache "Unprocessed_symbols" so we don't have to lock and unlock the mutex on every single iteration
    queue_length = Unprocessed_symbols;
    pthread_mutex_unlock(&Sym_mutex);

    for(i=0;i<queue_length;i++){
      if(Locked == -1 || interleaver_phase == Locked){
	// Assemble last N1 symbols from buffer, feed to Viterbi decoder
	for(b=0; b<N1; b++){
	  int offset;
	  
	  offset = N1 * permute(N1-1-b) + (N1-1-b);
	  symbol_buffer[b] = Decoder_Q[(Decoder_read - offset) & (DECODER_SIZE-1)];
	}
	update_viterbi27_blk(Decoders[interleaver_phase].vit,symbol_buffer,N1/2);
	
	// read out beginning of Viterbi decoder memory
	chainback_viterbi27(Decoders[interleaver_phase].vit,output,8*sizeof(output),-1);
	
	for(b=0;b<N1/16;b++){
	  // Write early part decoded data to output;
	  c = output[b];
	  // Output of Viterbi decoder is MSB first */
	  for(a=7;a>=0;a--){
	    process_hdlc_bit(interleaver_phase,(c >> a) & 1);
	  }
	}
      }
      Decoder_read++;
      interleaver_phase = (interleaver_phase + 1) % N1;
      Decoded_syms_processed++;
    }
    pthread_mutex_lock(&Sym_mutex);
    Unprocessed_symbols -= queue_length;
    if(Unprocessed_symbols <= DECODER_SIZE/2 - 1)
      pthread_cond_signal(&Sym_q_not_full); // If queue was full, unblock demodulator
    pthread_mutex_unlock(&Sym_mutex);
  }
  eof:;
  return NULL;
}

void symbolq(int s){
  assert(s >= 0 && s < 256);
  Histogram[s]++;
  Decoder_symbols_processed++;
  if(Verbose > 2 && (Decoder_symbols_processed % Histogram_interval) == 0){
    // Dump statistics and reset
    decoder_stats();
    memset(Histogram,0,sizeof(Histogram));
  }
  // Don't let the queue get more than halfway ahead, as the decoder will be looking
  // up to an interleaver span (N1*N1) back as it de-interleaves. Otherwise the demodulator could overwrite
  // the older symbols needed by the decoder.
  pthread_mutex_lock(&Sym_mutex);
  while(Unprocessed_symbols == DECODER_SIZE/2)
    pthread_cond_wait(&Sym_q_not_full,&Sym_mutex);

  Decoder_Q[Decoder_write & (DECODER_SIZE-1)] = s;
  Unprocessed_symbols++;
  if(Unprocessed_symbols > Highwater)
    Highwater = Unprocessed_symbols;
  if(Unprocessed_symbols == 1)
    pthread_cond_signal(&Sym_q_not_empty); // If queue was empty, signal the decoder
  pthread_mutex_unlock(&Sym_mutex);

  // Non-critical updates
  Decoder_write++;
}


void init_decoder(void){
  int k;
  int polys[2] = { V27POLYB, V27POLYA };

  if(Verbose)
    atexit(decoder_stats);

  // Initialize the queue between demodulator and decoder
  pthread_mutex_init(&Sym_mutex,NULL);
  pthread_cond_init(&Sym_q_not_empty,NULL);
  pthread_cond_init(&Sym_q_not_full,NULL);

  set_viterbi27_polynomial(polys);

  // Initialize decoder state for every interleaver phase
  memset(Decoders,0,sizeof(Decoders)); // clears outbuf and outbits
  for(k=0;k<N1;k++){
    Decoders[k].magic = DECODER_MAGIC;
    Decoders[k].state = ABORT; // Must see an initial flag
    // The minimum Viterbi decoder memory is driven by the following:
    // Chainbacks are done after each interleaver row (N1 = 128 symbols, 64 bits) has been
    // processed, so it must be at least 64 bits long or data will be lost before it is read out.

    // But it must be longer than that for the surviving paths to merge and give good BER performance.
    // The usual rule of thumb for that is an extra >= 128 bits between the oldest input symbol and the newest bit read in the chainback.
    // That gives 64 bits read out + 128 bits for the paths to merge = 192 bits, so we round up to 256.
    // Going a *lot* higher than this will only waste memory and increase latency.
    Decoders[k].vit = create_viterbi27(256); // Long enough for good performance
    Decoders[k].crc = 0xffffffff;
  }
}

// Process a single bit from the Viterbi decoder into the HDLC state machine
// Possible outcomes:
// 1. Nothing special (frame continues to be built)
// 2. HDLC abort, possibly continuing
// 3. Flag received: frame with good CRC
// 4. Flag received: frame with bad CRC
// 5. Flag received: frame is a runt
// 6. Frame gets too long

// We don't really know the correct interleaver phase until the first HDLC frame decodes successfully.
// So the decoder initially tries all N1 (128) interleaver phases in parallel.

// When a frame decodes, we can "lock" to that phase and stop decoding all the other phases.
// If any HDLC error (abort, bad CRC, runt, excessively long frame) then occurs to the expected phase,
// we have to consider the possibility that we've slipped in timing.
// So we revert to trying all interleaver phases until we again get a good frame.

// The user has a reliability/CPU time tradeoff option (-a) that forces the attempted decoding of 
// every interleaver phase all the time at the expense of extra CPU time.
// This probably ought to be the default since my Viterbi decoder is so fast,
// but the inelegancy still bugs me.

// I've tried code that backs up and retries all phases after losing lock,
// but it's a lot trickier than I thought so I've abandoned that for now.

// Maybe I should have included a sync pattern that could be easily found with a correlator, as I had
// done on AO-40. The overhead would only have been 1/128 = 0.78%.
unsigned int process_hdlc_bit(int phase,unsigned int bit){
  int length = 0;
  struct decoder *d = &Decoders[phase];

  assert(d->magic == DECODER_MAGIC);


  d->history = (d->history << 1) | bit;
  switch(d->history & 0xff){
  case 0x7e:
    // Received a complete flag
    // One of several things has happened:
    // 0. We've seen a flag after being in search mode (after an abort or excessively large frame)
    // 1. A good frame has been received
    // 2. A frame with a bad CRC has been received
    // 3. The frame is a runt with less than 4 bytes between flags
    //    since the CRC is 4 bytes.
    // (Null frames with CRC but no data are permitted, though not very useful.)
    if(d->state == ABORT || d->state == TOO_BIG){
      // We were searching for a flag, now we've found one
      // Strictly speaking these states are probably not necessary, as the CRC would undoubtedly fail anyway
      d->state = OK;
      return 0;
    }
    if(d->outbits >= 39){ // 32 bit CRC + first 7 bits of closing flag
      // The CRC is now immediately updated after every data bit, rather than on every byte,
      // so the first 7 bits of the closing flag have been clocked into the CRC to produce this
      // oddball CRC constant -- iff the CRC register was zero (correct) after the last bit of the incoming CRC.
      if((d->crc & 0xffffffff) == 0xec7dd02d){
	// Successful frame!
	d->state = OK;
	d->goodframes++;
	Goodframes++;
	length = (d->outbits >> 3); // includes CRC; rounds up if there's a fractional byte at the end
	Goodbytes += length - 4;
	
	if(Verbose > 1){
	  int days,hours,minutes;
	  double seconds;
	  char *lockmsg;
	  
	  lockmsg = (!All_phases && Locked == -1) ? "; Decoder locked\n" : "\n";
	  seconds = (double)Decoded_syms_processed / SYMRATE;
	  hms(&days,&hours,&minutes,&seconds);
	  pthread_mutex_lock(&Stderr_mutex);
	  fprintf(stderr,"decode-fec (%d:%02d:%02d:%06.3lf): phase %d, Good HDLC frame, %u bytes; Totals: frames %'llu; output bytes %'llu; input queue: %u syms%s",
		  days,hours,minutes,seconds,phase,length-4,
		  Goodframes,Goodbytes,Unprocessed_symbols,lockmsg);
	  pthread_mutex_unlock(&Stderr_mutex);
	}
	// Write decoded data to standard output in one of three formats:
	// By default, the frame length is written as 2 bytes, MSB first, followed by that many bytes of data
	// The CRC *IS* included.

	// If -H is set, frames are output in asynchronous HDLC format, including the original CRC-32.
	// Each frame starts and ends with a flag (0x7e)
	// Every occurrence of 0x7e or 0x7d inside a frame is replaced with the 2-byte sequence
	// 0x7d (escape) and the original character Xored with 0x20

	// If -C (clean) is set, each frame is output without any indication of its boundaries
	if(Hdlc){
	  // Send frame as asynch HDLC with original CRC
	  int i;
	  putchar(0x7e); // Opening flag
	  for(i=0;i<length;i++){ // include original CRC
	    if(d->buffer[i] == 0x7e || d->buffer[i] == 0x7d){
	      putchar(0x7d);
	      putchar(d->buffer[i] ^ 0x20);
	    } else {
	      putchar(d->buffer[i]);
	    }
	  }
	  putchar(0x7e); // closing flag
	} else {
	  if(!Clean){ // Emit frame length in big-endian order regardless of machine byte order
	    putchar((length >> 8) & 0xff);
	    putchar(length & 0xff);
	    fwrite(d->buffer,length,sizeof(char),stdout);
	  } else {
	    fwrite(d->buffer,length-4,sizeof(char),stdout);
	  }
	  fflush(stdout);
	}
	if(!All_phases)
	  Locked = phase; // Lock onto this phase, if enabled
      } else {	// End of frame, but CRC failed
	d->state = CRC_FAIL;
	if(DECODER_TEST || Locked != -1){
	  CRC_fails++;
	  if(Verbose > 1){
	    int days,hours,minutes;
	    double seconds;
	    
	    seconds = (double)Decoded_syms_processed / SYMRATE;
	    hms(&days,&hours,&minutes,&seconds);
	    
	    pthread_mutex_lock(&Stderr_mutex);
	    fprintf(stderr,"decode-fec (%d:%02d:%02d:%06.3lf): phase %d CRC failure %x; raw length %d bits\n",days,hours,minutes,seconds,phase,d->crc,d->outbits);
	    pthread_mutex_unlock(&Stderr_mutex);
	  }
	}
	Locked = -1; // Drop out of lock
      } // end of CRC validity checks
    } else if(d->outbits >= 8){ // d->outbits < 8 would be an idle flag stream, which we simply ignore.
      // We've received a runt, i.e., a frame with more than 1 but less than 4 bytes between flags.
      // Good frames always have at least 4 bytes since that is the length of the CRC.
      d->state = RUNT;
      if(DECODER_TEST || Locked != -1){
	Runts++;

	if(Verbose > 1){
	  int days,hours,minutes;
	  double seconds;
	  
	  seconds = (double)Decoded_syms_processed / SYMRATE;
	  hms(&days,&hours,&minutes,&seconds);
	  
	  pthread_mutex_lock(&Stderr_mutex);
	  fprintf(stderr,"decode-fec (%d:%02d:%02d:%06.3lf): phase %d runt; raw length %d bits\n",days,hours,minutes,seconds,phase,d->outbits);
	  pthread_mutex_unlock(&Stderr_mutex);
	}
      }
    }
    // Do this for all flags, regardless of whether the frame was good or not
    // Reset the HDLC state machine: CRC and data frame assembly
    memset(d->buffer,0,(d->outbits+7)/8);
    d->outbits = 0;
    d->crc = 0xffffffff;
    return length;    // End of flag processing
  case 0x7f:
    // Abort, toss everything
    d->state = ABORT;
    if(Locked != -1 || DECODER_TEST){
      Aborts++;
      if(Verbose > 1){
	int days,hours,minutes;
	double seconds;
	
	seconds = (double)Decoded_syms_processed / SYMRATE;
	hms(&days,&hours,&minutes,&seconds);
	pthread_mutex_lock(&Stderr_mutex);
	fprintf(stderr,"decode-fec (%d:%02d:%02d:%06.3lf): phase %d abort; raw length %d bits\n",days,hours,minutes,seconds,phase,d->outbits);
	pthread_mutex_unlock(&Stderr_mutex);
      }
    }
    Locked = -1; // Drop out of lock
    memset(d->buffer,0,(d->outbits+7)/8);
    d->outbits = 0;
    d->crc = 0xffffffff;
    return 0;
  case 0xff:
    // Abort continues, ignore
    return 0;
  case 0xfe:
    // Coming out of an abort; we're already waiting for a flag
    return 0;
  }
  // Reach here on anything not a flag or abort
  if((d->history & 0x3f) == 0x3e){
    // Stuffed 0-bit after 5 1-bits, ignore
    return 0;
  }
  // See if frame has gotten too big
  if(d->outbits == 8*MAX_HDLC_SIZE){
    // Too big, treat like an abort
    if(DECODER_TEST || Locked != -1){
      Too_big++;
      if(Verbose > 1){
	int days,hours,minutes;
	double seconds;
	
	seconds = (double)Decoded_syms_processed / SYMRATE;
	hms(&days,&hours,&minutes,&seconds);
	
	pthread_mutex_lock(&Stderr_mutex);
	fprintf(stderr,"decode-fec (%d:%02d:%02d:%06.3lf): phase %d too-long frame; raw length %d bits\n",days,hours,minutes,seconds,phase,d->outbits);
	pthread_mutex_unlock(&Stderr_mutex);
      }
    }
    d->state = TOO_BIG;
    memset(d->buffer,0,(d->outbits+7)/8);
    d->outbits = 0;
    d->crc = 0xffffffff;
    Locked = -1; // Drop out of lock
    return 0;
  }
  // Ordinary data bit processing
  // This requires that d->buffer[] be cleared after every reset!
  if(bit)
    d->buffer[d->outbits>>3] |= 1 << (d->outbits & 7);
  
  // Update CRC
  {
    int feedback;
    feedback = ((d->crc >> 31) ^ bit) & 1;
    d->crc <<= 1;
    if(feedback)
      d->crc ^= CRCPOLY;
  }
  d->outbits++;
  return 0;
}

void decoder_stats(void){
  int i;

  pthread_mutex_lock(&Stderr_mutex);
  fprintf(stderr,"decode-fec: Symbols processed: %'lld; decoded bytes: %'lld; decoded frames %'lld; queue length %'u; queue high water %'u\n",
	  Decoder_symbols_processed,Goodbytes,Goodframes,Unprocessed_symbols,Highwater);

  fprintf(stderr,"   Runts %d; Too big %d; Aborts %d; CRC failures %d; good frames %lld\n",
	  Runts, Too_big, Aborts, CRC_fails, Goodframes);
  fprintf(stderr,"decode-fec: Viterbi decoder input histogram:\n");
  for(i=0;i<256;i++){
    if((i % 16) == 0)
      fprintf(stderr,"%3d:",i);
    fprintf(stderr,"%8lld",Histogram[i]);
    if((i % 16) == 15)
      fprintf(stderr,"\n");
  }
  pthread_mutex_unlock(&Stderr_mutex);
}
