// Non-coherent BPSK demod and decoder for Suitsat-2
// Multithreaded version, parallel searches the signal space until decoding succeeds
// June 2010, Phil Karn, KA9Q
// Copyright 2010, Phil Karn. This software may be used under the terms
// of the GNU General Public License (GPL).
#define _FILE_OFFSET_BITS 64
#define _FORTIFY_SOURCE 2

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <limits.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include <pthread.h>
#include <sched.h>
#include <locale.h>
#include <time.h>
#include "dspfec.h"
#include "dpsk.h"

// SYMPCHUNK sets the number of symbols demodulated at once with a constant carrier frequency and symbol timing offset.
// This establishes the equivalent frequency and time tracking bandwidth
// This is a fundamental decoder parameter that establishes its carrier and symbol timing tracking bandwidth.
// Read the separate file "sympchunk.txt" for notes on the significance of this value and how it affects the demodulator's ability to
// accurately estimate carrier frequency and symbol timing in noisy signals and to track changes in those parameters due to Doppler and oscillator frequency offsets.
#define SYMPCHUNK (4 * N1)
#define NSEARCHERS 4                          // Maximum number of simultaneous search/demod jobs. Must be at least 3. Ideally a power of 2, but not required
#define SAMPPCHUNK (SYMPCHUNK * SAMPPSYM)     // samples in one chunk
#define SAMPBUFSIZE (2 * N1 * N1 * SAMPPSYM)  // Size of sample buffer: 32.768 sec, two interleaver spans

#if ((SAMPBUFSIZE % SYMPCHUNK) != 0)
#error "SAMPBUFSIZE must be integral multiple of SYMPCHUNK"
#endif

// Searcher used for brute-force carrier frequency and symbol timing searches
struct searcher_state {
  pthread_t thread;                  // Thread descriptor for job
  double energy;                     // Total demodulator output energy (output)
  signed short *samples;             // Pointer to start of samples to be processed by this job (input)
  unsigned int cphase_inc_start;     // First carrier frequency to try (phase increment / sample, 2^32 = 2 pi radians) (input)
  unsigned int cphase_inc_step;      // Increment of carrier frequency to try (phase increment / sample, 2^32 = 2 pi radians) (input)
  unsigned int nfreq;                // Number of frequencies to search, starting at cphase_inc_start (input)
  unsigned int cphase_inc;           // Optimum carrier frequency increment (output)
  int symphase;                      // Optimum Symbol timing offset (output)
};


// DBPSK demodulator I/O
struct demod_state {
  int symphase;                // Offset to first sample of first symbol; determined by symbol timing (input)
  double energy;               // Total demodulator output energy (output)
  long long data[SYMPCHUNK];  // Demodulated symbols (output)
  int symbol_count;
};

struct searcher_state Searchers[NSEARCHERS];


struct option Options[] = {
  { "all-phases",         0, NULL, 'a' },
  { "gain",               1, NULL, 'g' },
  { "verbose",            0, NULL, 'v' },
  { "carrier",            1, NULL, 'c' },
  { "carrier-range",      1, NULL, 'r' },
  { "histogram-interval", 1, NULL, 'i' },
  { "frame",              0, NULL, 'f' },
  { "locale",             1, NULL, 'l' },
  { "disable-SIMD",       0, NULL, 'd' },
  { "no-SIMD",            0, NULL, 'd' },
  { "search-interval",    1, NULL, 's' },
  { "clean",              0, NULL, 'C' },
  { "hdlc",               0, NULL, 'H' },
  { "version",            0, NULL, 'V' },
  { NULL,                 0, NULL,  0  }
};

// Input sample buffer and related metadata
signed short *Samples;
unsigned long long Unprocessed_samples;
int EOF_encountered;        // signals end of file on stream
int EOD_encountered;        // signals end of demodulated data to decoder
pthread_mutex_t Samp_mutex; // Mutex and condition used for locking Samples buffer when reading from stream
pthread_cond_t Samp_not_full;
pthread_cond_t Samp_not_empty;

pthread_mutex_t Stderr_mutex; // Mutex on stderr to keep diag messages from jumbling up

pthread_attr_t Searcher_attr;

int File_mode;                 // 1 = reading from memory mapped file, 0 = reading from stream via circular buffer
int Input;                     // Input file descriptor
void *Dp;                      // matched filter correlator control block
long long Sampcounter;         // Total samples read by demodulator
long long Samples_read;        // Total samples read by reader
long long Symbols_demodulated; // Total symbols demodulated

// Defaults: search 1200, 1300, 1400, 1500, 1600, 1700, 1800 Hz
double Carrier = 1500;         // Center of carrier frequency search range
double Carrier_range = 300;    // Limits of search range above and below Carrier frequency
int Search_interval;
int All_phases;                // Decode all deinterleaver phases; don't try to lock onto a single phase
int Clean;                     // Output data without lengths; default is to send length + data
int Hdlc;                      // Output data as asynch HDLC (see PPP, RFC1662) with original CRC32

unsigned int Cphase_inc_start;
unsigned int Cphase_inc_stop;
unsigned int Cphase_inc_step;
unsigned int Ftotal;
unsigned int Fperslot;
double Gain = 128; // Heuristically, this seems about optimum
  
pthread_t Reader_thread;
pthread_t Decoder_thread;

struct demod_state Demod_state[3];

// Declarations of thread functions
void *reader(void *);
void *searcher(void *);
int demodulate(signed short *,signed short *,int,double *,long long *);
double freq_tracker(signed short *,int,unsigned int);
void init_decoder(void);

extern int Histogram_interval;
extern int Framing_mode;
extern short FIR_taps[];
extern int Ntaps;

int main(int argc, char *argv[]){
  int option_index, i, mode;
  char *filename, *locale;
  struct stat statbuf;
  signed short baseband_i[SAMPPCHUNK], baseband_q[SAMPPCHUNK], *read_ptr;
  int symphase;
  unsigned int symbol_count, cphase, cphase_inc, cphase_tmp, samples_processed;
  unsigned long long chunk;
  double carrier;
  int best_index;
  size_t stacksize;
  long long *data;

  Histogram_interval = SYMPCHUNK; // Default
  locale = "en_US.UTF-8";
  Search_interval = 16;

  // Process command line options
  while((i = getopt_long(argc, argv, "as:dg:fvc:r:i:l:CHV", Options, &option_index)) != EOF){
    switch(i){
    case 'a':
      All_phases = 1;
      break;
    case 'd':
      Cpu_features = NONE; // Disable any SIMD vector processing (MMX,SSE, Altivec/VMX, etc)
      break;
    case 'g':
      Gain = atof(optarg);
      break;
    case 'f':
      Framing_mode = 1;
      break;
    case 'v':
      Verbose++;
      break;
    case 'c':
      Carrier = atof(optarg);
      break;
    case 'r':
      Carrier_range = atof(optarg);
      break;
    case 's':
      Search_interval = atoi(optarg);
      break;
    case 'i':
      Histogram_interval = atoi(optarg); // Number of symbols between statistics dumps
      break;
    case 'l':
      locale = optarg;
      break;
    case 'C':
      Clean++;
      break;
    case 'H':
      Hdlc++;
      break;
    case 'V':
      fprintf(stderr,"BPSK1000 demodulator by Phil Karn, KA9Q, copyright 2010; built on %s at %s\n",__DATE__,__TIME__);
      exit(0);
      break;
    }
  }
  cpu_id();
  if(setlocale(LC_ALL, locale) == NULL){
    fprintf(stderr, "setlocale(%s) failed\n", locale);
  }
  Input = 0; // Default input is stdin

  init_sin();

  Cphase_inc_start = (Carrier - Carrier_range) * E32 / SAMPRATE;
  Cphase_inc_stop =  (Carrier + Carrier_range) * E32 / SAMPRATE;
  Cphase_inc_step = 100. * E32 / SAMPRATE; // Search carrier frequency in 100 Hz steps
  Ftotal = 2 * Carrier_range/100. + 1;
  Fperslot = (Ftotal+NSEARCHERS-1)/NSEARCHERS;

  if(Verbose){
    fprintf(stderr, "Nominal carrier %.1lf; minimum carrier %.1lf; maximum carrier %.1lf\n", Carrier, Carrier-Carrier_range, Carrier+Carrier_range);
    fprintf(stderr, "Frequencies to search: %u; Search slots: %d; Frequencies per search slot: %u\n", Ftotal, NSEARCHERS, Fperslot);
  }

  // Scale FIR taps so that the sum of their absolute values is less than 64K
  // This prevents overflow of a 32-bit dot product with worst case input
  {
    
    unsigned int i;
    double sum,scale_factor;

    sum = 0;
    for(i=0;i<Ntaps;i++)
      sum += abs(FIR_taps[i]);

    if(sum > 65536){
      scale_factor = 65536./(double)sum;
      for(i=0;i<Ntaps;i++)
	FIR_taps[i] *= scale_factor;
      if(Verbose)
	fprintf(stderr, "Scaling FIR taps by %lg\n", scale_factor);
    }
  }

  // Set up receive filter
  if((Dp = initdp(FIR_taps,Ntaps+1)) == NULL){
    fprintf(stderr, "%s: Can't set up matched filter\n", argv[0]);
    exit(1);
  }

  // Initialize decoder before starting it - this avoids a race
  init_decoder();

  // Set up attributes for the searcher threads, ensuring enough room on stack for baseband I/Q data
  pthread_attr_init(&Searcher_attr);
  pthread_attr_getstacksize(&Searcher_attr,&stacksize);
  stacksize += 2 * sizeof(signed short) * SAMPPCHUNK;
  pthread_attr_setstacksize(&Searcher_attr,stacksize);
  
  pthread_mutex_init(&Stderr_mutex, NULL);

  pthread_create(&Decoder_thread, NULL, decoder, NULL);

  if(optind < argc){
    // File name argument is present, read from it instead of stdin
    filename = argv[optind];
    Input = open(filename, O_RDONLY);
    if(Input == -1){
      fprintf(stderr, "Can't read %s: %s\n", filename,strerror(errno));
      exit(1);
    }
  } else {
    filename = "[stdin]";
  }

  // See if we are reading from a file, either specified as an argument or redirected into stdin.
  // If so, try to mmap it into memory
  // If mmap fails, revert to streaming mode
  if(fstat(Input,&statbuf) == -1){
    fprintf(stderr, "fstat of %s failed: %s\n", filename, strerror(errno));
    exit(1);
  }
  mode = statbuf.st_mode & S_IFMT;
  if(Verbose){
    switch(mode){
    case S_IFIFO:
      fprintf(stderr, "Input is a FIFO\n");
      break;
    case S_IFCHR:
      fprintf(stderr, "Input is a character special device\n");
      break;
    case S_IFDIR:
      fprintf(stderr, "error: Input is a directory\n");
      break;
    case S_IFBLK:
      fprintf(stderr, "Input is a block special device\n");
      break;
    case S_IFREG:
      fprintf(stderr, "Input is a regular file\n");
      break;
    case S_IFLNK:
      fprintf(stderr, "Input is a symbolic link\n");
      break;
    case S_IFSOCK:
      fprintf(stderr, "Input is a socket\n");
      break;
#ifdef S_IFWHT
    case S_IFWHT:
      fprintf(stderr, "error: Input is a whiteout\n");
      break;
#endif
    default:
      fprintf(stderr, "Input type is unknown: %d 0%o 0x%x\n", mode, mode, mode);
      exit(1);
    }
  }
  if(mode == S_IFREG){
    // Input is regular file
    File_mode++;
    if(Verbose){
      int days,hours,minutes;
      double seconds;

      seconds = (double)statbuf.st_size / (SAMPRATE * sizeof(signed short));
      hms(&days,&hours,&minutes,&seconds);

      fprintf(stderr,"file contains %'llu bytes; %'llu samples; %'llu chunks; %d days %02d:%02d:%06.3lf @ %'d Hz sample rate\n",
	      (long long unsigned)statbuf.st_size,
	      (long long unsigned)statbuf.st_size / sizeof(signed short),
	      (long long unsigned)statbuf.st_size / (SAMPPCHUNK * sizeof(signed short)),
	      days,hours,minutes,seconds,SAMPRATE);
    }
    Samples = NULL;
    Unprocessed_samples = statbuf.st_size / sizeof(signed short);
    // Ask kernel to map entire sample file into any place in memory it chooses, read only
    Samples = mmap(&Samples, statbuf.st_size, PROT_READ, MAP_FILE|MAP_SHARED, Input, (off_t)0);
    if(Samples == MAP_FAILED){
      fprintf(stderr, "memory mapping of %s failed: %s; reverting to stream mode\n", filename, strerror(errno));
      File_mode = 0;
      Unprocessed_samples = 0;
      Samples = NULL;
    }
#ifdef S_IFWHT
  } else if(mode == S_IFDIR || mode == S_IFWHT || mode == S_IFLNK){
#else
  } else if(mode == S_IFDIR || mode == S_IFLNK){
#endif
    fprintf(stderr,"Can't handle input file type 0%o\n", mode);
    exit(1);
  }


  // These won't be used if we're reading from a memory-mapped file
  if(!File_mode){
    void *p;

    pthread_mutex_init(&Samp_mutex, NULL);
    pthread_cond_init(&Samp_not_full, NULL);
    pthread_cond_init(&Samp_not_empty, NULL);

    // Allocate stream buffer with one chunk overrun area
    p = malloc(sizeof(signed short) * (SAMPBUFSIZE+SAMPPCHUNK));
    Samples = (signed short *)p;
    if(Samples == NULL){
      fprintf(stderr,"Can't malloc streaming buffer\n");
      exit(1);
    }
    pthread_create(&Reader_thread, NULL, reader, NULL);    // Start reader
    pthread_detach(Reader_thread);
  }


  // These keep lint happy
  symbol_count = 0;
  symphase = 0;
  carrier = 0;
  cphase = 0;
  cphase_inc = 0;

  read_ptr = Samples;

  for(chunk=0;;chunk++){
    double energy,gain,rms,lower,upper;
    int i;
    double cpt0,cpt1;
    struct searcher_state *s;
    double input_rms = 0;
    

    if(File_mode){
      if(Unprocessed_samples < SAMPPCHUNK)
	goto eof;
    } else {
      // Wait for at least one chunk of samples to arrive, if necessary
      pthread_mutex_lock(&Samp_mutex);
      while(Unprocessed_samples < SAMPPCHUNK){
	// Wait for at least one chunk to arrive
	if(EOF_encountered){
	  // Reached end of file
	  pthread_mutex_unlock(&Samp_mutex);
	  goto eof;
	}
	pthread_cond_wait(&Samp_not_empty,&Samp_mutex);
      }
      pthread_mutex_unlock(&Samp_mutex);
    }
#if 0
    for(i=0;i<SAMPPCHUNK;i++)
      assert(read_ptr[i] != -32768);
#endif
    if(Verbose)
      input_rms = sqrt((double)sumsq(read_ptr,SAMPPCHUNK)/SAMPPCHUNK);

    // Perform brute force search every N chunks
    if((chunk % Search_interval) == 0){
      // Spawn threads to search signal space
      // This needs to be done more carefully, to avoid unnecessary symbol slips in corner cases
      unsigned int slot;
      double maxenergy_value;
      int slots;
      int fleft;

      cphase_inc = Cphase_inc_start;
      fleft = Ftotal;
      for(slot=0;fleft > 0 && slot < NSEARCHERS;slot++){
	s = &Searchers[slot];
	s->samples = read_ptr;
	s->cphase_inc_start = cphase_inc;
	s->cphase_inc_step = Cphase_inc_step;
	s->nfreq = min(Fperslot,fleft);
	cphase_inc += Cphase_inc_step * s->nfreq;
	fleft -= s->nfreq;
	pthread_create(&s->thread,&Searcher_attr,searcher,s);
      }
      // Find the winner
      slots = slot;
      maxenergy_value = 0;
      for(slot=0;slot < slots;slot++){
	s = &Searchers[slot];
	pthread_join(s->thread,NULL);
	if(s->energy >= maxenergy_value){
	  maxenergy_value = s->energy;
	  cphase_inc = s->cphase_inc;
	  symphase = s->symphase;
	}
      }

      carrier = (double)cphase_inc * SAMPRATE / E32;
      if(symphase < SAMPPSYM/2)
	symphase += SAMPPSYM; // Allow room for early symbol timing test

      if(Verbose > 1){
	pthread_mutex_lock(&Stderr_mutex);
	fprintf(stderr,"   full search: carrier %.1lf Hz; best offset %2d; energy %.0lg\n",
		carrier,
		symphase,
		maxenergy_value);
	pthread_mutex_unlock(&Stderr_mutex);
      }
    }

    // track frequency with two probes, one below and one above nominal carrier frequency
    // Use linear interpolation to determine more accurate carrier frequency
    lower = carrier - 75;
    upper = carrier + 75;
    cpt0 = freq_tracker(read_ptr,symphase, (unsigned int) (lower * E32 / SAMPRATE));
    cpt1 = freq_tracker(read_ptr,symphase, (unsigned int) (upper * E32 / SAMPRATE));

    // Solve for zero crossing of cpt / carrier curve
    // Note: this isn't exact. The cpt values aren't linear as they're actually functions of the sine of the frequency offset
    // I should work up a more exact function, or at least bound the error
    if(cpt1 != cpt0) // Avoid unlikely divide by zero
      carrier = lower - cpt0 * (upper - lower) / (cpt1 - cpt0);

    // Spin down to baseband with best estimate of carrier frequency
    cphase_inc = carrier * E32 / SAMPRATE;
    cphase_tmp = cphase;
    for(i=0; i < SAMPPCHUNK; i++){
      baseband_i[i] = ((int)read_ptr[i] * (int)cos_table(cphase_tmp)) >> 15;
      baseband_q[i] = ((int)read_ptr[i] * (int)sin_table(cphase_tmp)) >> 15;
      cphase_tmp += cphase_inc;
    }	
    // Demodulate 3 times: early, on time, and late
    best_index = -1;
    energy = 0;
    for(i=0;i<3;i++){
      Demod_state[i].symbol_count = demodulate(baseband_i, baseband_q, symphase + i - 1, &Demod_state[i].energy, Demod_state[i].data);
      if(energy < Demod_state[i].energy){
	best_index = i;
	energy = Demod_state[i].energy;
      }
    }
    assert(best_index != -1);
    symphase = symphase + best_index - 1;
    symbol_count = Demod_state[best_index].symbol_count;
    data = Demod_state[best_index].data;
    Symbols_demodulated += symbol_count;

    // Scale demodulated data and pass to FEC
    rms = sqrt(energy / symbol_count);
    gain = Gain / rms; // Tune this
    // Skip data[0], which is actually the last symbol from the previous chunk
    // It was needed by the demodulator as the reference for differentially decoding the second symbol, i.e., data[1]
    for(i=1;i<symbol_count;i++){
      double y;
      
      y = 127.5 + gain * data[i];
      y = (y > 255) ? 255 : ((y < 0) ? 0 : y);
      
      symbolq((int)y); // Pass to FEC decoder thread
    }
    // Move carefully to next chunk, allowing for overlap of last/first symbol used for differential decoding and for timing skew
    samples_processed = SAMPPSYM * (symbol_count-1);
    if(symphase <= SAMPPSYM/2){
      // Timing is moving early; move back 1/2 symbol
      samples_processed -= SAMPPSYM/2;
      symphase += SAMPPSYM/2;
    } else if(symphase >= (3*SAMPPSYM)/2){
      // Timing is moving late; move forward 1/2 sample
      samples_processed += SAMPPSYM/2;
      symphase -= SAMPPSYM/2;
    }
    if(Verbose > 1){
      int days,hours,minutes;
      double seconds;
      
      seconds = (double)chunk*SAMPPCHUNK/SAMPRATE;
      hms(&days,&hours,&minutes,&seconds);

      pthread_mutex_lock(&Stderr_mutex);
      fprintf(stderr,"demod-dpsk (%d:%02d:%02d:%06.3lf): input rms %.0lf; carrier %.1lf; symphase %d\n",
	      days,hours,minutes,seconds,input_rms,carrier,symphase);
      pthread_mutex_unlock(&Stderr_mutex);	
    }

    Sampcounter += samples_processed;
    cphase += samples_processed * cphase_inc;
    read_ptr += samples_processed;

    if(File_mode){
      Unprocessed_samples -= samples_processed;
    } else {
      if(read_ptr >= &Samples[SAMPBUFSIZE])
	read_ptr -= SAMPBUFSIZE; // Wrap around circular buffer used only in streaming mode
  
      // Tell reader thread that we've read another chunk
      pthread_mutex_lock(&Samp_mutex);
      Unprocessed_samples -= samples_processed;
      pthread_cond_signal(&Samp_not_full); // Wake up reader, if it's waiting for more space to appear
      pthread_mutex_unlock(&Samp_mutex);
    }
  }
 eof:;

  for(i=0;i<N1+N1*N1;i++)
    symbolq(128); // Flush interleavers and Viterbi decoder with erasures
  Symbols_demodulated += N1+N1*N1;
  EOD_encountered++;
  pthread_join(Decoder_thread,NULL);

  exit(0);
}  

// Repeatedly demodulate at a specified carrier frequency for all symbol offsets
// record one that gives greatest demodulated energy
void *searcher(void *arg){
  struct searcher_state *s = (struct searcher_state *)arg;
  signed short baseband_i[SAMPPCHUNK], baseband_q[SAMPPCHUNK];
  unsigned int cphase_inc;
  int j;

  s->energy = 0;
  s->symphase = -1; // to keep compiler happy about uninitialized variables
  for(j=0,cphase_inc = s->cphase_inc_start; j < s->nfreq; j++,cphase_inc += s->cphase_inc_step){
    int i,offset;
    unsigned int cphase;

    // Downconvert chunk of samples to baseband 
    cphase = 0;
    for(i=0; i < SAMPPCHUNK; i++){
      baseband_i[i] = ((int)s->samples[i] * (int)cos_table(cphase)) >> 15;
      baseband_q[i] = ((int)s->samples[i] * (int)sin_table(cphase)) >> 15;
      cphase += cphase_inc;
    }	
    
    // Perform demodulation for all possible symbol timings
    // looking for maximum energy
    for(offset=0; offset < SAMPPSYM; offset++){ // For every symbol timing offset
      int bb_p;
      int tlast_i,tlast_q;
      double energy;
      
      tlast_i = 0;
      tlast_q = 0;
      
      energy = 0;
      for(bb_p = offset; bb_p + Ntaps < SAMPPCHUNK; bb_p += SAMPPSYM){ // For every symbol in baseband buffer
	int fi,fq;
	long long symbol;
	
	// Demodulate through matched filter, compute baseband energy
	fi = dotprod(Dp, &baseband_i[bb_p]);
	fq = dotprod(Dp, &baseband_q[bb_p]);
	
	// Dot product of previous and current center complex samples gives differentially demodulated symbol
	symbol = (long long)fi * tlast_i + (long long)fq * tlast_q;
	
	energy += (double)symbol * symbol;
	tlast_i = fi;
	tlast_q = fq;
      }
      // Keep track of frequency and timing that gives maximum energy
      if(energy >= s->energy){ // Guarantee at least one will be taken
	s->cphase_inc = cphase_inc;
	s->energy = energy;
	s->symphase = offset;
      }
    }
  }
  pthread_exit(NULL);
}
// Demodulate chunk, determining frequency error
double freq_tracker(signed short *samples,        // Pointer to start of samples to be processed by this job (input)
		    int symphase,                 // Offset to first sample of first symbol; determined by symbol timing (input)
		    unsigned int cphase_inc){     // Carrier phase increment per sample (2^32 = 2 pi radians) (input)

  signed short baseband_i[SAMPPCHUNK],baseband_q[SAMPPCHUNK];
  unsigned int cphase;
  int bb_p,i,tlast_i,tlast_q;
  double cpt;

  // Downconvert chunk of samples to baseband 
  cphase = 0;
  for(i=0; i < SAMPPCHUNK; i++){
    baseband_i[i] = ((int)samples[i] * (int)cos_table(cphase)) >> 15;
    baseband_q[i] = ((int)samples[i] * (int)sin_table(cphase)) >> 15;
    cphase += cphase_inc;
  }
  // Perform demodulation with specified symbol timing, computing sum of cross products
  cpt = 0;
  tlast_i = tlast_q = 0;
  for(bb_p = symphase; bb_p + Ntaps < SAMPPCHUNK; bb_p += SAMPPSYM){ // For every symbol in baseband buffer
    int fi,fq;
    long long symbol;
    
    // Demodulate through matched filter, compute baseband energy
    fi = dotprod(Dp, &baseband_i[bb_p]);
    fq = dotprod(Dp, &baseband_q[bb_p]);
    
    // Dot product of previous and current center complex samples gives differentially demodulated symbol
    symbol = (long long)fi * tlast_i + (long long)fq * tlast_q;
    // When there hasn't been a transition, a positive cross product means the signal frequency is positive, i.e.,
    // our estimate of carrier frequency is too low. A transition reverses the sense of the cross product.
    if(symbol < 0)
      cpt -= (long long)fi * tlast_q - (long long)fq * tlast_i;
    else
      cpt += (long long)fi * tlast_q - (long long)fq * tlast_i;
    
    tlast_i = fi;
    tlast_q = fq;
  }
  return cpt;
}

// DBPSK demodulate chunk, return symbol count
int demodulate(signed short *baseband_i,      // Baseband I-channel samples (input)
		 signed short *baseband_q,    // Baseband Q-channel samples (input)
		 int symphase,                // Offset to first sample of first symbol; determined by symbol timing (input)
		 double *energy,              // Total demodulator output energy (output)
                 long long *data){            // Demodulated symbols, SAMPPCHUNK entries, allocated by caller (output)

  int bb_p, symbol_count = 0, tlast_i = 0, tlast_q = 0;
  double en_tmp;

  en_tmp = 0;
  for(bb_p = symphase; bb_p + Ntaps < SAMPPCHUNK; bb_p += SAMPPSYM){ // For every symbol in baseband buffer
    int fi, fq;
    long long symbol;
    
    // Sample matched filter
    fi = dotprod(Dp, &baseband_i[bb_p]);
    fq = dotprod(Dp, &baseband_q[bb_p]);
    
    // Dot product of previous and current complex samples gives differentially demodulated symbol
    // data[0] will always be 0
    symbol = (long long)fi * tlast_i + (long long)fq * tlast_q;
    en_tmp += (double)symbol * symbol;

    assert(symbol_count < SYMPCHUNK);
    data[symbol_count++] = symbol;
    tlast_i = fi;
    tlast_q = fq;
  }
  if(energy != NULL)
    *energy = en_tmp;
  return symbol_count;
}

// Reader thread - reads audio samples into circular buffer, wakes up demodulator threads
//  Not started when reading from a mapped-in file
void *reader(void *arg){
  int remain;
  signed short *write_ptr;
  unsigned char *p;

  write_ptr = Samples;
  while(1){
    assert(write_ptr <= &Samples[SAMPBUFSIZE]);
    // Wait for sufficient room to appear in circular buffer
    pthread_mutex_lock(&Samp_mutex);
    while(Unprocessed_samples >= SAMPBUFSIZE/2)
      pthread_cond_wait(&Samp_not_full, &Samp_mutex);

    pthread_mutex_unlock(&Samp_mutex);

    // Read SAMPPCHUNK symbols, looping in case a read temporarily comes up short
    p = (unsigned char *)write_ptr;
    remain = SAMPPCHUNK * sizeof(signed short); // bytes remaining to be read
    while(remain != 0){
      int cnt;

      cnt = read(Input,p,remain);
      if(cnt <= 0){
	if(cnt < 0){
	  pthread_mutex_lock(&Stderr_mutex);	  
	  fprintf(stderr, "Input read error: %s\n",strerror(errno));
	  pthread_mutex_unlock(&Stderr_mutex);
	}
	break; // EOF or error; quit
      }
      p += cnt;
      remain -= cnt;
    }
    if(remain != 0)
      break;      // Error or eof encountered; ignore any partial chunk

    clip(write_ptr, SAMPPCHUNK, SHRT_MAX, -SHRT_MAX); // Clip any -32768 samples to -32767

    // If we've written the overrun area, copy it to the beginning of the buffer and wrap around, skipping it
    if(write_ptr == &Samples[SAMPBUFSIZE]){
      memmove(Samples, write_ptr, SAMPPCHUNK * sizeof(signed short));
      write_ptr = &Samples[SAMPPCHUNK];
    } else {
      // Otherwise simply increment the write pointer
      write_ptr += SAMPPCHUNK;
    }
    Samples_read += SAMPPCHUNK;

    // Wake up demodulator waiting for new data to appear
    pthread_mutex_lock(&Samp_mutex);
    Unprocessed_samples += SAMPPCHUNK;
    pthread_cond_signal(&Samp_not_empty);
    pthread_mutex_unlock(&Samp_mutex);
  }
  // EOF on input; append enough zeroes to flush interleaver, wait for buffer to empty, then exit
  if(Verbose){
#if (_POSIX_TIMERS > 0) && _POSIX_THREAD_CPUTIME // Doesn't seem to be on any system I have
    struct timespec tp;
    double cpu_seconds;
    clock_gettime(CLOCK_THREAD_CPUTIME_ID,&tp);
    cpu_seconds = tp.tv_sec + tp.tv_nsec / 1.e9;
    pthread_mutex_lock(&Stderr_mutex);
    fprintf(stderr, "Reader: read EOF after %'llu samples, exiting, CPU time = %.3lf sec\n", Samples_read,cpu_seconds);
    pthread_mutex_unlock(&Stderr_mutex);
#else
    pthread_mutex_lock(&Stderr_mutex);
    fprintf(stderr, "Reader: read EOF after %'llu samples, exiting\n", Samples_read);
    pthread_mutex_unlock(&Stderr_mutex);
#endif
  }
  // Alert demodulator that an EOF has been seen
  pthread_mutex_lock(&Samp_mutex);
  EOF_encountered = 1;
  pthread_cond_signal(&Samp_not_empty);
  pthread_mutex_unlock(&Samp_mutex);
  return NULL;
}
