/* 16-bit signed integer dot product
 * portable C version
 *
 * Copyright 2009 Phil Karn
 * May be used under the terms of the GNU public license
 *
 * $Id$
 */

#ifndef NULL
#define NULL ((void *)0)
#endif

#include <stdlib.h>
#include <assert.h>
#include "dspfec.h"

void *initdp_p(signed short *,int);
void *initdp_sse2(signed short *,int);
void *initdp_mmx(signed short *,int);
void *initdp_av(signed short *,int);

void freedp_p(void *);
void freedp_mmx(void *);
void freedp_sse2(void *);
void freedp_av(void *);

long long dotprodl_p(void *,signed short *);
long long dotprodl_mmx(void *,signed short *);
long long dotprodl_sse2(void *,signed short *);
long long dotprodl_sse41(void *,signed short *);

int dotprod_p(void *,signed short *);
int dotprod_mmx(void *,signed short *);
int dotprod_sse2(void *,signed short *);
int dotprod_av(void *,signed short *);


void *initdp(signed short coeffs[],int len){
  if(Cpu_features == UNKNOWN)
    cpu_id();

  switch(Cpu_features){
#if __i386__ || __i486__ || __i586__ || __i686__ || __x86_64__
  case SSE42:
  case SSE41:
  case SSSE3:
  case SSE3:
  case SSE2:
    return initdp_sse2(coeffs,len);
  case SSE:
  case MMX:
    return initdp_mmx(coeffs,len);
#elif __ppc__
  case ALTIVEC:
    return initdp_av(coeffs,len);
#endif
  default:
    return initdp_p(coeffs,len);
  }
}

void freedp(void *p){
  switch(Cpu_features){
#if __i386__ || __i486__ || __i586__ || __i686__ || __x86_64__
  case SSE42:
  case SSE41:
  case SSSE3:
  case SSE3:
  case SSE2:
    freedp_sse2(p);
    break;
  case SSE:
  case MMX:
    freedp_mmx(p);
    break;
#elif __ppc__
  case ALTIVEC:
    freedp_av(p);
    break;
#endif
  default:
    freedp_p(p);
    break;
  }
}

long long dotprodl(void *p,signed short a[]){
  switch(Cpu_features){
#if __i386__ || __i486__ || __i586__ || __i686__ || __x86_64__
  case SSE42:
  case SSE41:
    return dotprodl_sse41(p,a);
  case SSSE3:
  case SSE3:
  case SSE2:
    return dotprodl_sse2(p,a);
  case SSE:
  case MMX:
    return dotprodl_mmx(p,a);
#endif
  default:
    return dotprodl_p(p,a);
  }
}


int dotprod(void *p,signed short a[]){
  switch(Cpu_features){
#if __i386__ || __i486__ || __i586__ || __i686__ || __x86_64__
  case SSE42:
  case SSE41:
  case SSSE3:
  case SSE3:
  case SSE2:
    return dotprod_sse2(p,a);
  case SSE:
  case MMX:
    return dotprod_mmx(p,a);
#elif __ppc__
  case ALTIVEC:
    return dotprod_av(p,a);
#endif
  default:
    return dotprod_p(p,a);
  }
}

/* Internal definitions for dotproduct function */
struct dotprod_p {
  int len; /* Number of coefficients */

  /* On a SSE2 machine, these hold 8 copies of the coefficients,
   * preshifted by 0,1,...7 words to meet all possible input data
   * alignments (see Intel ap559 on MMX dot products).
   *
   */
  signed short *coeffs;
};


/* Create and return a descriptor for use with the dot product function */
void *initdp_p(signed short coeffs[],int len){
  struct dotprod_p *dp;
  int j;

  if(len <= 0)
    return NULL;

  dp = (struct dotprod_p *)calloc(1,sizeof(struct dotprod_p));
  dp->len = len;

 /* Just one copy of the coefficients for the C version */
 dp->coeffs = (signed short *)malloc(len*sizeof(signed short));
 for(j=0;j<len;j++)
   dp->coeffs[j] = coeffs[j];

  return (void *)dp;
}

/* Free a dot product descriptor created earlier */
void freedp_p(void *p){
  struct dotprod_p *dp = (struct dotprod_p *)p;

  free(dp->coeffs);
  free(dp);
}

/* Compute a dot product given a descriptor and an input array
 * The length is taken from the descriptor
 */
long long dotprodl_p(void *p,signed short a[]){
  struct dotprod_p *dp = (struct dotprod_p *)p;
  long long corr = 0;
  int i;

  for(i=0;i<dp->len;i++){
    corr += (long long)a[i] * dp->coeffs[i];
  }
  return corr;
}

int dotprod_p(void *p,signed short a[]){
  struct dotprod_p *dp = (struct dotprod_p *)p;
  int corr = 0;
  int i;

  for(i=0;i<dp->len;i++){
    corr += (int)a[i] * dp->coeffs[i];
  }
  return corr;
}
