#if !defined(__SSE2__)
#error "Must compile with -msse2 or better"
#endif

/* 16-bit signed integer dot product using SSE-2 or SSE-4.1
 * Copyright 2009 Phil Karn
 * May be used under the terms of the GNU public license
 *
 * $Id$
 */
#include <assert.h>
#include "sse.h"
#include "dotprod_sse2.h"

/* Create and return a descriptor for use with the dot product function */
void *initdp_sse2(signed short coeffs[],int len){
  struct dotprod_sse2 *dp;
  int j,i;
  signed short *wp;

  if(len <= 0)
    return NULL;

  dp = (struct dotprod_sse2 *)calloc(1,sizeof(struct dotprod_sse2));
  dp->len = len;

  /*
    1->8*8:  10000000; 01000000; 00100000; 00010000; 00001000; 00000100; 00000010; 00000001;
    2->9*8:  12000000; 01200000; 00120000; 00012000; 00001200; 00000120; 00000012; 00000001 20000000;
    6->13*8: 12345600; 01234560; 00123456; 00012345 60000000; 00001234 56000000; 00000123 45600000; 00000012 34560000; 00000001 23456000;
    
    8->15*8: 12345678; 01234567 80000000; 00123456 78000000; 00012345 67800000; 00001234 56780000; 00000123 45678000; 00000012 34567800; 00000001 23456780; 
    9->16*8: 12345678 90000000; 01234567 89000000; 00123456 78900000; 00012345 67890000; 00001234 56789000; 00000123 45678900; 00000012 34567890; 00000001 23456789;
    10->17*8:12345678 9A000000; 01234567 89A00000; 00123456 789A0000; 00012345 6789A000; 00001234 56789A00; 00000123 456789A0; 00000012 3456789A; 00000001 23456789 A0000000; 
  */
  /* Allocate aligned space for 8 copies of the coefficients, shifted by 0-7 words */
   wp = MALLOC(sizeof(signed short)*(len+7)*8);
   assert(((long)wp & 0xf) == 0);
   //   memset(wp,0,sizeof(signed short)*(len+7)*8);
   for(i=0;i<8;i++){
     dp->coeffs[i] = (v8hi *)wp;
     for(j=0;j<i;j++)
       *wp++ = 0; /* Leading zeroes; wp += i */
     for(j=0;j<len;j++)
       *wp++ = coeffs[j];
     while((long)wp & 0xf)
       *wp++ = 0;     /* Trailing zeroes; wp += 7 - (7+len+i) % 8 */
   }
   assert(wp <= (signed short *)dp->coeffs[0] + (len+7)*8); /* Ensure we haven't overflowed */
   return (void *)dp;
}

/* Free a dot product descriptor created earlier */
void freedp_sse2(void *p){
  struct dotprod_sse2 *dp = (struct dotprod_sse2 *)p;

  free(dp->coeffs[0]);
  free(dp);
}

/* Compute a dot product given a descriptor and an input array
 * The length is taken from the descriptor
 * This SSE-assisted version returns 32 bits, so it's faster than the 64-bit version but overflows are silently discarded
 * To avoid them, scale your coefficients so that the sum of their absolute values is < 65535
 */

int dotprod_sse2(void *p,signed short a[]){
  struct dotprod_sse2 *dp = (struct dotprod_sse2 *)p;
  int al,i,vcnt;
  v8hi *ar,*coeffs;
  union {
    int d[4]; /* 32 bits */
    v4si v;
  } sums;
  v4si s;

  /* Round down to 16-byte boundary. We'll operate
   * on data before the user's buffer, but the corresponding
   * coefficients are zero so there's no harm. A segfault can't occur
   * either because the page boundaries are much larger than 16 bytes
   */
  ar = (v8hi *)((long)a & ~15);
  al = (short *)a - (short *)ar; /* Number of words misaligned */
  
  vcnt = (dp->len+al+7)/8; /* Number of vectors, rounding up to include partial ones */
  coeffs = dp->coeffs[al];
  
  /* Do 8 signed 16-bit multiply-adds per loop. Each pair of multiply-adds is combined into
   * one of four 32-bit sums.
   */
  s = _mm_madd_epi16(coeffs[0],ar[0]);                   /* [A6*B6+A7*B7]   [A4*B4+A5*B5]  [A2*B2+A3*B3]    [A0*B0+A1*B1] */
  for(i=1;i<vcnt;i++)
    s = _mm_add_epi32(s,_mm_madd_epi16(coeffs[i],ar[i]));                   /* [A6*B6+A7*B7]   [A4*B4+A5*B5]  [A2*B2+A3*B3]    [A0*B0+A1*B1] */

  /* return the sum of the four 32-bit sub sums */
  sums.v = s;
  return sums.d[0] + sums.d[1] + sums.d[2] + sums.d[3];
}


/* SSE2-assisted dotproduct, returns 64 bits
 * This one can't overflow regardless of arguments as long as the vectors are less than 2^32 elements
 * But this comes at a significant cost in performance, so use the 32-bit version with scaled coefficients if possible
 */
long long dotprodl_sse2(void *p,signed short a[]){
  struct dotprod_sse2 *dp = (struct dotprod_sse2 *)p;
  int al,i,vcnt;
  v8hi *ar,*coeffs;
  v2di sums; /* two 64-bit sums */
  union {
    long long s[2];
    v2di v;
  } s;

  /* Round down to 16-byte boundary. We'll operate
   * on data before the user's buffer, but the corresponding
   * coefficients are zero so there's no harm. A segfault can't occur
   * either because the page boundaries are much larger than 16 bytes
   */
  ar = (v8hi *)((long)a & ~15);
  al = (short *)a - (short *)ar; /* Number of words misaligned */
  
  vcnt = (dp->len+al+7)/8; /* Number of vectors, rounding up to include partial ones */
  sums = _mm_setzero_si128();
  coeffs = dp->coeffs[al];
  
  /* Do 8 signed 16-bit multiply-adds per loop. Each pair of multiply-adds is combined into
   * one of four 32-bit sums. This is just wide enough for one pair, so to prevent overflow
   * we immediately extend each to 64 bits and add into two 64-bit running sub-sums.
   */
  for(i=0;i<vcnt;i++){
    v4si isums;
    isums = _mm_madd_epi16(coeffs[i],ar[i]);                   /* [A6*B6+A7*B7]   [A4*B4+A5*B5]  [A2*B2+A3*B3]    [A0*B0+A1*B1] */

    /* SSE4.1 has a 32-64-bit sign extend, but SSE2 does not.
     * But could this actually be slower??
     */
    /* SSE2 doesn't have a 32->64 bit sign extend, so we have to do it ourselves.
     * The 32 sign bits will be 0 for positive values and all 1's for negative values.
     */
    {
      v4si sign;
      sums += _mm_unpacklo_epi32(isums,_mm_setzero_si128());     /* [0]              [A2*B2+A3*B3] [0]              [A0*B0+A1*B1] */
      sign = _mm_unpacklo_epi32(_mm_setzero_si128(),isums);      /* [A2*B2+A3*B3]    [0]           [A0*B0+A1*B1]    [0]           */
      sums += _mm_srai_epi32(sign,31);                           /* [<0 ? 1's : 0]   [0]           [<0 ? 1's : 0]   [0]          .*/ 
      
      sums += _mm_unpackhi_epi32(isums,_mm_setzero_si128());     /* [0]              [A6*B6+A7*B7] [0]              [A4*B4+A5*B5] */
      sign = _mm_unpackhi_epi32(_mm_setzero_si128(),isums);      /* [A6*B6+A7*B7]    [0]           [A4*B4+A5*B5]    [0]           */
      sums += _mm_srai_epi32(sign,31);                           /* [<0 ? 1's : 0]   [0]           [<0 ? 1's : 0]   [0]           */ 
    }
  }
  /* return the sum of the two 64-bit sub sums */
  sums += _mm_unpackhi_epi64(sums,_mm_setzero_si128());
  s.v = sums;
  return s.s[0];
}
