// Basic constants for the ARISSAT-1 BPSK-1000 telemetry format
// July 2010, Phil Karn, KA9Q
// Copyright 2010, Phil Karn. This software may be used under the terms of the GNU Public License (GPL)

// Constants defined by the telemetry format (i.e., the air interface)
#define SYMRATE    1000                // Symbol rate, Hz (must be divisible into SAMPRATE) 
#define N1         128                 // Dimension of convolutional interleaver 
#define CRCPOLY    0x04c11db7          // CRC-32 polynomial; high order term in msb
// hex:       0    4    c   1    1     d    b    7
// binary: 0000 0100 1100 0001 0001 1101 1011 0111
// Here's the CRC generator polynomial in 4-term groups, one per hex nybble
// The exponent of x corresponds to the 1-bit in the polynomial word
// x^32 (implicit)
//                        (hex 0)
//  + x^26                (hex 4)
//  + x^23 + x^22         (hex c)
//  + x^16                (hex 1)
//  + x^12                (hex 1)
//  + x^11 + x^10 + x^8   (hex d)
//  + x^7 + x^5 + x^4     (hex b)
//  + x^2 + x + 1         (hex 7)

// convolutional encoder constants for r=1/2 k=7 encoder
// The NASA-DSN convention sends V27POLYA inverted, then V27POLYB
// The CCSDS/NASA-GSFC convention sends V27POLYB, then V27POLYA inverted
// They invert alternate symbols to ensure a minimum channel symbol transition density for receiver
// symbol tracking in case a long string of 1's or 0's is transmitted.

// We don't need that here because we use HDLC and differential PSK (and we don't need compatibility).
// Our differential encoding represents a '1' as no phase transition and a '0' as an inversion (180 deg shift)
// so a long string of 0's produces continuous transitions.
// HDLC stuffs a 0 bit after every 5 1's in a data stream, so a long string of 1's is no problem either.

// Essentially our HDLC uses the NRZI (non return to zero, inverted) mode already standard in amateur packet radio
// except for:
// not using the AX-25 header;
// using a 32-bit CRC instead of a 16-bit CRC;
// following the HDLC encoder with a k=7 r=1/2 convolutional encoder and N=128 bit-reversed convolutional interleaver;
// and using differential binary phase shift keying (DBPSK) modulation
// instead of the usual (and EXTREMELY wasteful) Bell 202 modem (noncoherent BFSK with 1200 and 2200 Hz tones) on a FM main carrier.

#define ICRATE     2              // Inverse of convolutional encoder rate 1/2
#define CK         7              // Convolutional encoder constraint length 
#define	V27POLYA   0x6d           // First symbol generator polynomial:  1101101 binary, 155 octal
#define	V27POLYB   0x4f           // Second symbol generator polynomial: 1001111 binary, 117 octal
// Many references show these two polynomials in opposite bit order, i.e.,
// G1 = 1011011 = 133 octal
// G2 = 1111001 = 171 octal
// so we're using the very same code even though it might not seem so!

// End of format-defined constants

// This establishes the sample rate for the audio A/D. This could be a user-controlled variable, but
// there's not really much need for it. Virtually all computer sound A/Ds can do 48 kHz, and
// a high sample rate is desirable to permit precise symbol tracking. It also permits use of the
// same FIR filter coefficients as on the spacecraft, which also operates at 48 kHz.
#define SAMPRATE   48000               // Audio A/D sample rate, Hz
// If another rate is chosen, it MUST be an integer multiple of the symbol rate, and you must
// redo the filter coefficients in nyquist.c
// If you *really* need to use another sample rate, it's probably easiest to just convert it to 48 kHz
// with a utility like 'sox'. It can also handle fractional sample rates, useful to correct for small clock errors.

// Derived rate constants, here to avoid forward references:
#define SAMPPSYM   (SAMPRATE/SYMRATE)  // Samples per symbol, i.e., 48
#define BITRATE    (SYMRATE/ICRATE)    // i.e., 500 bps

void putsym(int);
void putbit(int);
void symbolq(int);
void *decoder(void *); // Decoder thread
