/* Generate Suitsat-2/ARISSat-1 telemetry:
   CRC-32 / HDLC / r=1/2 k=7 convolutional / 128way interleaver / DBPSK modulator
 * Input is arbitrary data, interpreted LSB first (HDLC standard)
 * Output is 48 kHz 16-bit linear PCM little-endian (regardless of machine order!)
 *
 * Copyright 2010 Phil Karn, KA9Q
 * May be used under the terms of the GNU General Public License
 */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <assert.h>
#include "dspfec.h"
#include "dpsk.h"

/* Defaults */
int Framesize = 128;

struct option Options[] = {
  { "frame-size",1,NULL,'f' },
  { "verbose", 0, NULL, 'v' },
  { "doppler-rate",1,NULL,'r' },
  { "doppler-limit",1,NULL,'l' },
  { "carrier",1,NULL,'c' },
  { "test-message",0,NULL,'t' },
  { 0, 0, NULL, 0}
};

unsigned long long Totalbits = 0;
unsigned long long Frames = 0;

extern int Samprate;
extern int Symrate;
extern int Ntaps;

/* External variables set by fir_taps.c */
extern short FIR_taps[];
extern int Samprate;
extern int Symrate;
extern int Ntaps;

struct { // Modulator state
  unsigned int cphasestep;
  unsigned int cphase;
  int fir;
  int lastbit;
  double nominal_carrier;
  double current_carrier;
  double Doppler_rate;
  double Doppler_limit;
  enum { UP, DOWN} Doppler_direction;
} Modulator;  

struct encoder_state {
  int sr;
  int interleaver[2*N1][N1];
  int i;
  int j;
};

struct encoder_state Encoder;

#define FLAG 0x7e

int send_hdlc(unsigned char *,int);
void putbit(int);
void putsym(int);

int main(int argc,char *argv[]){
  int i,len,option_index;
  unsigned char *buffer;
  FILE *fp,*fp_null;
  int test_mode = 0;
  
  Modulator.nominal_carrier = 1500;
  while((i = getopt_long(argc,argv,"f:vr:l:c:t",Options,&option_index)) != EOF){
    switch(i){
    case 'v':
      Verbose++;
      break;
    case 'f': /* Size of each data frame but the last */
      Framesize = atoi(optarg);
      break;
    case 'c':
      Modulator.nominal_carrier = atof(optarg);
      break;
    case 'r':
      Modulator.Doppler_rate = atof(optarg);
      break;
    case 'l':
      Modulator.Doppler_limit = atof(optarg);
      break;
    case 't':
      test_mode = 1;
      break;
    case '?':
      fprintf(stderr,"Usage: %s [--carrier|-c <nominal carrier frequency>] [--doppler_rate|-r <doppler rate, Hz/sec] [--doppler_limit|-l <Doppler limit, Hz>] [[-v] [-v]]\n",argv[0]);
      exit(1);
      break;
    }
  }
  cpu_id();
  
  if(Verbose){
    fprintf(stderr,"%s: Center carrier %.1lf Doppler rate %.1lf Hz/sec, Doppler limit +/- %.1lf Hz;",argv[0],Modulator.nominal_carrier,Modulator.Doppler_rate,Modulator.Doppler_limit);
    fprintf(stderr," Sample rate %d Symbol rate %d\n",Samprate,Symrate);
  }

  buffer = malloc(Framesize);
  assert(buffer != NULL);
	  
  if(argc <= optind){
    fp = stdin;
  } else {
    if((fp = fopen(argv[optind],"r")) == NULL){
      perror("open");
      exit(1);
    }
  }
  init_sin(); /* shift 32-bit word right 24 bits to get index */
  Modulator.current_carrier = Modulator.nominal_carrier;
  Modulator.cphasestep = E32 * Modulator.current_carrier / Samprate;
  Modulator.Doppler_direction = DOWN;

  // Temporarily redirect output to /dev/null while we fill the interleaver with an encoded flag stream
  fp_null = fopen("/dev/null","w");
  if(fp_null != NULL){
    FILE *stdout_save;
    stdout_save = stdout;
    stdout = fp_null;

    // Fill interleaver with flags
    for(i=0;i<(N1*N1/16);i++)
      for(len=7;len>=0;len--)
	putbit((FLAG >> len) & 1);

    stdout = stdout_save;
    fclose(fp_null);
  }
  if(test_mode){
    // Enter infinite test loop
    for(i=0;;i++){
      unsigned char buffer[1000];
      snprintf((char *)buffer,sizeof(buffer),"Test frame %d\n",i);
      Frames++;
      Totalbits += send_hdlc(buffer,strlen((char *)buffer));
    }
  } else {
    while((len = fread(buffer,1,Framesize,fp)) > 0){
      Frames++;
      Totalbits += send_hdlc(buffer,len);
    }
    fclose(fp);
  }
  // Flush interleaver, twice, with flags
  for(i=0;i<2048;i++)
    for(len=7;len>=0;len--)
      putbit((FLAG >> len) & 1);

  if(Verbose){
    fprintf(stderr,"%s: Frames sent: %llu; bits sent: %llu\n",argv[0],Frames,Totalbits);
  }
  exit(0);
}

/* Send HDLC frame, CRC and closing flag. We assume a flag was just sent; we could
 * send one to make sure but that would be wasteful, especially with small frames
 * Return count of total bits sent
 */
int send_hdlc(unsigned char *buffer,int len){
  int bits,onebits,bit;
  unsigned int crc;

  bits = len * 8;
  onebits = 0;
  crc = 0xffffffff;
  while(len-- != 0){
    unsigned char c;
    int j,feedback;
    
    c = *buffer++;
    for(j=0;j<8;j++){ /* For each byte of data, sent LSB first */
      bit = (c >> j) & 1;
      // Update CRC-32
      feedback = ((crc & 0x80000000) >> 31) ^ bit;
      crc = crc << 1;
      if(feedback)
	crc ^= CRCPOLY;
      
      putbit(bit);
      if(bit){
	if(++onebits >= 5){
	  bits++;
	  putbit(0);
	  onebits = 0;
	}
      } else
	onebits = 0;
    }
  }
  /* Emit CRC here */
  for(len=31;len>=0;len--){
    /* CRC is sent high bit first, reverse of data */
    bit = (crc >> len) & 1;
    putbit(bit);
    if(bit){
      if(++onebits >= 5){
	bits++;
	putbit(0);
	onebits = 0;
      }
    } else
      onebits = 0;
  }
  // Closing flag
  for(len=7;len>=0;len--)
    putbit((FLAG >> len) & 1);
  return bits + 40; // data bits + stuffed bits + CRC-32 + flag
}

/* Convolutional encode, convolutional interleave */

inline int permute(int j){
  return bitreversen(j,N1);
}

void putbit(int bit){
  int row;

  /* Send bytes LSBs first to be consistent with HDLC encoder convention */
  Encoder.sr = (Encoder.sr<<1) | bit;

  /* write  before read */
  Encoder.interleaver[Encoder.i][Encoder.j] = parityb(Encoder.sr & V27POLYB);
  row = (Encoder.i - permute(Encoder.j) + 2*N1) % (2*N1);
    
  putsym(Encoder.interleaver[row][Encoder.j]);
  
  Encoder.j++; // This can't wrap if N1 is even

  Encoder.interleaver[Encoder.i][Encoder.j] = parityb(Encoder.sr & V27POLYA);
  row = (Encoder.i - permute(Encoder.j) + 2*N1) % (2*N1);
  putsym(Encoder.interleaver[row][Encoder.j]);

  if(++Encoder.j == N1){
    Encoder.j = 0;

    if(++Encoder.i == 2*N1)
      Encoder.i = 0;
  }
}

void putsym(int sym){
  int j;

  /* Differential BPSK, LSB first: 0 means a change in carrier phase */
  Modulator.lastbit ^= (sym & 1) ^ 1;
  
  Modulator.fir = Modulator.fir << 1 | Modulator.lastbit;
  // Generate each sample within the current symbol
  for(j=0;j<SAMPPSYM;j++){
    int fout,tmp,i;
    signed short sample;
    
    fout = 0;
    tmp = Modulator.fir;
    for(i=Ntaps-1-j;i>=0;i-= SAMPPSYM){
      fout += tmp & 1 ? FIR_taps[i] : -FIR_taps[i];
      assert(fout >= -32767 && fout <= 32767);
      tmp >>= 1;
    }
    sample = (fout * cos_table(Modulator.cphase)) >> 16;
    if(Modulator.Doppler_rate != 0 && Modulator.Doppler_limit != 0){	  /* Doppler adjustment */
      if((Modulator.Doppler_direction == UP) & (Modulator.current_carrier >= Modulator.nominal_carrier + Modulator.Doppler_limit)){
	/* Reached upper limit */
	Modulator.Doppler_direction = DOWN;
	if(Verbose)
	  fprintf(stderr,"Nominal carrier %.1lf\n",Modulator.current_carrier);
	
      } else if(Modulator.Doppler_direction == DOWN && (Modulator.current_carrier <= Modulator.nominal_carrier - Modulator.Doppler_limit)){
	/* Reached lower limit */
	Modulator.Doppler_direction = UP;
	if(Verbose)
	  fprintf(stderr,"Nominal carrier %.1lf\n",Modulator.current_carrier);
      }
      if(Modulator.Doppler_direction == UP){
	Modulator.current_carrier += Modulator.Doppler_rate / Samprate;	    /* Go up */
      } else {
	Modulator.current_carrier -= Modulator.Doppler_rate / Samprate;	    /* Go down */
      }
      Modulator.cphasestep = E32 * Modulator.current_carrier / Samprate;
    }
    Modulator.cphase += Modulator.cphasestep;
    fwrite(&sample,sizeof(sample),1,stdout); // Write in machine order
  } /* For each sample in a bit */
}

