#include "dspfec.h"

void ll2int_p(signed int *,signed long long *,int);
void int2short_p(signed short *,signed int *,int);
void short2char_p(signed char *,signed short *,int);
void char2short_p(signed short *,signed char *,int);
void short2int_p(signed int *,signed short *,int);
void int2ll_p(signed long long *,signed int *,int);

// Beginning of SSE2 vectorized versions - the compact versions are especially useful since they saturate
#include "sse.h"
#if !defined(__SSE2__)
#error "Must compile with -msse2 or better"
#endif

// Pack signed 64-bit to 32 bit
void ll2int_sse2(signed int *out,signed long long *in,int cnt){
  v4si *outp;
  v2di *inp;
  int i,n;

  // Pass unaligned block off to portable routine
  // It's possible to pre-process certain combinations of unaligned pointers but it's too hard right now.
  // Easier to just ensure that callers provide aligned arrays
  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    ll2int_p(out,in,cnt);
    return;
  }
  outp = (v4si *)out;
  inp = (v2di *)in;

  // Use SIMD for 4 elements or more
  n = cnt / 4; // number of 4x32-bit output vectors
  for(i=0; i<n; i++)
    outp[i] = _mm_packs_epi32(inp[2*i],inp[2*i+1]);

  cnt -= n*8;

  // Handle unaligned trailer
  if(cnt != 0){
    out += n*8;
    in += n*8;
    ll2int_p(out,in,cnt);
  }
}

// Pack signed 32-bit to 16 bit
void int2short_sse2(signed short *out,signed int *in,int cnt){
  v8hi *outp;
  v4si *inp;
  int i,n;

  // Pass unaligned block off to portable routine
  // It's possible to pre-process certain combinations of unaligned pointers but it's too hard right now.
  // Easier to just ensure that callers provide aligned arrays
  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    int2short_p(out,in,cnt);
    return;
  }
  outp = (v8hi *)out;
  inp = (v4si *)in;

  // Use SIMD for 8 elements or more
  n = cnt / 8; // number of 8x16-bit output vectors
  for(i=0; i<n; i++)
    outp[i] = _mm_packs_epi32(inp[2*i],inp[2*i+1]);

  cnt -= n*8;

  // Handle unaligned trailer
  if(cnt != 0){
    out += n*8;
    in += n*8;
    int2short_p(out,in,cnt);
  }
}
// Pack 16-bit short to 8-bit char
void short2char_sse2(signed char *out,signed short *in,int cnt){
  v16qi *outp;
  v8hi *inp;
  int i,n;

  // Pass unaligned block off to portable routine
  // It's possible to pre-process certain combinations of unaligned pointers but it's too hard right now.
  // Easier to just ensure that callers provide aligned arrays
  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    short2char_p(out,in,cnt);
    return;
  }
  outp = (v16qi *)out;
  inp = (v8hi *)in;

  // Use SIMD for 16 elements or more
  n = cnt / 16; // number of 16 x8-bit output vectors
  for(i=0; i<n; i++)
    outp[i] = _mm_packs_epi16(inp[2*i],inp[2*i+1]);

  cnt -= n*16;

  // Handle unaligned trailer
  if(cnt != 0){
    out += n*16;
    in += n*16;
    short2char_p(out,in,cnt);
  }
}

// Unpack 8-bit signed char to 16-bit signed short with sign extension
void char2short_sse2(signed short *out,signed char *in,int cnt){
  v16qi *outp;
  v16qi *inp;
  v16qi mask;
  int i,n;

  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    char2short_p(out,in,cnt);
    return;
  }
  outp = (v16qi *)out;
  inp = (v16qi *)in;
  n = cnt / 16;
  for(i=0; i<n; i++){
    mask = _mm_cmpgt_epi8(_mm_set1_epi8(0),inp[i]); // mask = ff if inp[i] is negative, 0 if positive
    outp[2*i] = _mm_unpacklo_epi8(inp[i],mask);
    outp[2*i+1] = _mm_unpackhi_epi8(inp[i],mask);
  }
  cnt -= n*16;
  if(cnt != 0){
    out += n*16;
    in += n*16;
    char2short_p(out,in,cnt);
  }
}

// Unpack 16-bit signed short to 32-bit signed int with sign extension
void short2int_sse2(signed int *out,signed short *in,int cnt){
  v8hi *outp;
  v8hi *inp;
  v8hi mask;
  int i,n;

  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    short2int_p(out,in,cnt);
    return;
  }
  outp = (v8hi *)out;
  inp = (v8hi *)in;
  n = cnt / 8;
  for(i=0; i<n; i++){
    mask = _mm_cmpgt_epi16(_mm_set1_epi16(0),inp[i]); // mask = ffff if inp[i] is negative, 0 if positive
    outp[2*i] = _mm_unpacklo_epi16(inp[i],mask);
    outp[2*i+1] = _mm_unpackhi_epi16(inp[i],mask);
  }
  cnt -= n*8;
  if(cnt != 0){
    out += n*8;
    in += n*8;
    short2int_p(out,in,cnt);
  }
}

// Unpack 32-bit signed int to 64-bit signed long long with sign extension
void int2ll_sse2(signed long long *out,signed int *in,int cnt){
  v4si *outp;
  v4si *inp;
  v4si mask;
  int i,n;

  if(((long)out & 0xf) != 0 || ((long)in & 0xf) != 0){
    int2ll_p(out,in,cnt);
    return;
  }
  outp = (v4si *)out;
  inp = (v4si *)in;
  n = cnt / 4;
  for(i=0; i<n; i++){
    mask = _mm_cmpgt_epi32(_mm_set1_epi32(0),inp[i]); // mask = ffff if inp[i] is negative, 0 if positive
    outp[2*i] = _mm_unpacklo_epi32(inp[i],mask);
    outp[2*i+1] = _mm_unpackhi_epi32(inp[i],mask);
  }
  cnt -= n*4;
  if(cnt != 0){
    out += n*4;
    in += n*4;
    int2ll_p(out,in,cnt);
  }
}
