// This file is the same as viterbi27_sse2.c except for the function name,
// and the requirement that it be compiled with the -msse4.1 flag to invoke the alternate
// code that uses SSE4.1 instructions.

#include <assert.h>
#include <limits.h>
#include "dspfec.h"
#include "sse.h"

int V27_init;

#if !defined(__SSE4_1__)
#error "must compile with -msse2 or better"
#endif 

#include "viterbi27_sse2.h"


// Adapted from my r=1/6 k=15 decoder, which uses 16-bit path metrics
int update_viterbi27_blk_sse41(void *p,unsigned char *syms,int nbits){
  struct v27sse2 *vp = p;
  decision_sse2_t *d = vp->dp;
  int normalizes = 0;

  assert(vp != NULL && vp->magic == V27MAGIC);
  while(nbits--){
    __m128i sym0v,sym1v;
    void *tmp;
    int i;

    /* Splat the 0th symbol across sym0v, the 1st symbol across sym1v, etc */
    sym0v = _mm_set1_epi16(syms[0]);
    sym1v = _mm_set1_epi16(syms[1]);
    syms += 2;

    /* SSE2 doesn't support saturated adds on unsigned shorts, so we have to use signed shorts */
    for(i=0;i<4;i++){
      v8hi decision0,decision1,metric,m_metric,m0,m1,m2,m3,survivor0,survivor1;

      /* Form branch metrics
       * Because Branchtab takes on values 0 and 255, and the values of sym?v are offset binary in the range 0-255,
       * the XOR operations constitute conditional negation.
       * metric and m_metric (-metric) are in the range 0-511
       */
      metric = _mm_add_epi16(_mm_xor_si128(Branchtab27_sse2[0].v[i],sym0v),_mm_xor_si128(Branchtab27_sse2[1].v[i],sym1v));
      m_metric = _mm_sub_epi16(_mm_set1_epi16(511),metric); // Negate

      // Add branch metrics to path metrics to produce contending new path metrics
      // In the unlikely (impossible?) event of an overflow, the saturating adds turns an almost certain decoding error
      // into a tiny BER vs Eb/No degradation
      // The metrics are signed, ranging from -32768 to +32767
      m0 = _mm_adds_epi16(vp->old_metrics->v[i],metric);
      m1 = _mm_adds_epi16(vp->old_metrics->v[4+i],m_metric);
      m2 = _mm_adds_epi16(vp->old_metrics->v[i],m_metric);
      m3 = _mm_adds_epi16(vp->old_metrics->v[4+i],metric);

      decision0 = _mm_cmpgt_epi16(m0,m1);
      decision1 = _mm_cmpgt_epi16(m2,m3);

      // Signed compare and select available on SSE2; the unsigned version requires SSE4.1
      survivor0 = _mm_min_epi16(m0,m1);
      survivor1 = _mm_min_epi16(m2,m3);

      // Pack each set of decisions into 8 8-bit bytes, interleave and compress to 16 bits
      // The second argument to each _mm_packs_epi16 is a don't care, as they're excluded by the unpacklo_epi8()
      d->w[i] = _mm_movemask_epi8(_mm_unpacklo_epi8(_mm_packs_epi16(decision0,decision0),_mm_packs_epi16(decision1,decision1)));

      // Store surviving metrics
      vp->new_metrics->v[2*i] = _mm_unpacklo_epi16(survivor0,survivor1);
      vp->new_metrics->v[2*i+1] = _mm_unpackhi_epi16(survivor0,survivor1);
    }
    // See if we need to renormalize
    // This only occurs every few hundred bits so its speed isn't important
    if(vp->new_metrics->w[0] >= SHRT_MAX-12750){
      int i,mn;
      __m128i minv;
      
      // Find smallest metric
      minv = vp->new_metrics->v[0];
      for(i=1;i<8;i++)
	minv = _mm_min_epi16(minv,vp->new_metrics->v[i]);

      // This was originally a runtime test, i.e., if(Cpu_features >= SSE41) but it trapped at runtime on a non-SSE4.1 CPU when compiled with gcc -O2
      // The reason was that the optimizer speculatively executed the relatively cheap horizontal minpos instruction before waiting for the result of the compare.
      // This was a difference of opinion between me and the compiler on what constitutes "correct execution"; the compiler didn't know that the reason for the test was to avoid
      // an illegal instruction trap!
#if __SSE4_1__
      minv = _mm_minpos_epu16(minv); // horizontal minimum
#else
      minv = _mm_min_epi16(minv,_mm_srli_si128(minv,8));
      minv = _mm_min_epi16(minv,_mm_srli_si128(minv,4));
      minv = _mm_min_epi16(minv,_mm_srli_si128(minv,2));
#endif
      mn = _mm_extract_epi16(minv,0);
      minv = _mm_set1_epi16(mn - SHRT_MIN);

      /* We cannot use a saturated subtract, because we often have to adjust by more than SHRT_MAX
       * This is okay since it can't overflow anyway
       */
      for(i=0;i<8;i++)
	vp->new_metrics->v[i] = _mm_sub_epi16(vp->new_metrics->v[i],minv);
      normalizes++;
    }
    // Wrap around decision memory if necessar
    if(++d >= (decision_sse2_t *)(vp + 1)  + vp->length)
       d = (decision_sse2_t *)(vp + 1);

    /* Swap pointers to old and new metrics */
    tmp = vp->old_metrics;
    vp->old_metrics = vp->new_metrics;
    vp->new_metrics = tmp;
  }
  vp->dp = d;
  return normalizes;
}
