/* Test viterbi decoder in streaming mode
 * Look at error rates vs chainback distance, etc.
 * April 2009, Phil Karn, KA9Q
 * $Id$
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <memory.h>
#include <getopt.h>
#include "dspfec.h"

struct option Options[] = {
  {"frame-length",1,NULL,'l'},
  {"frame-count",1,NULL,'n'},
  {"ebn0",1,NULL,'e'},
  {"gain",1,NULL,'g'},
  {"verbose",0,NULL,'v'},
  {"impulse",0,NULL,'i'},
  {NULL},
};

#define RATE (1./2.)
#define MAXBYTES 10000

int Gain = 32;
int Impulse = 0;
int Framebits = 128;
int Bits = 10000;

#define BUFFERSIZE 2
#define DECODE_BUF 16

void rstats(void);

int main(int argc,char *argv[]){
  int bit,d,i;
  int sr=0;

  unsigned char symbols[BUFFERSIZE];
  int sym_p = 0;

  void *vp;
  extern char *optarg;
  double esn0,ebn0;
  double noise;
  long long totalerrors_static = 0;
  long long totalerrors_best = 0;

  long long bit_error_counts_0[DECODE_BUF*8];
  long long bit_error_counts_b[DECODE_BUF*8];

  unsigned char transmitted_data[DECODE_BUF+1];

  memset(bit_error_counts_0,0,sizeof(bit_error_counts_0));
  memset(bit_error_counts_b,0,sizeof(bit_error_counts_b));
  memset(transmitted_data,0,sizeof(transmitted_data));

  srandom(time(NULL));
  zigset(time(NULL));
  ebn0 = 3.0;
  while((d = getopt_long(argc,argv,"l:n:e:g:vi",Options,NULL)) != EOF){
    switch(d){
    case 'l':
      Framebits = atoi(optarg);
      break;
    case 'n':
      Bits = atoi(optarg);
      break;
    case 'e':
      ebn0 = atof(optarg);
      break;
    case 'g':
      Gain = atoi(optarg);
      break;
    case 'v':
      Verbose++;
      break;
    case 'i':
      Impulse++;
      break;
    }
  }
  cpu_id();
  if((vp = create_viterbi27(Framebits)) == NULL){
    printf("create_viterbi27 failed\n");
    exit(1);
  }

  esn0 = ebn0 + 10*log10((double)RATE); /* Es/No in dB */
  /* Compute noise voltage. The 0.5 factor accounts for BPSK seeing
   * only half the noise power, and the sqrt() converts power to
   * voltage.
   */
  noise = Gain * sqrt(0.5/pow(10.,esn0/10.));
  
  printf("Bits = %u\n",Bits);
  printf("Decoder memory = %u ebn0 = %.2f dB\n",Framebits,ebn0);
  printf("noise = %g RMS, data amplitude = +/-%d\n",noise,Gain);
  
  for(bit=0;bit<Bits;bit++){
    int databit,p;

    unsigned char decode_data_0[DECODE_BUF];
    unsigned char decode_data_b[DECODE_BUF];
    int error0,errorb,errors;
    int random_data = 0;
    int rbitsleft = 0;
    
    /* Need more data */
    if(!Impulse){
      if(rbitsleft == 0){
	/* random() has 31 random bits, with range of 0 to 0x7fffffff */
	random_data = random();
	rbitsleft = 31;
      }
      databit = random_data & 1;
      random_data >>= 1;
      rbitsleft--;
    } else 
      databit = ((bit % 256) == 0) ? 1 : 0; /* Impulse every 64 bits */
    
    /* Convolutional encode, add noise */
    sr = (sr << 1) | databit;
    symbols[sym_p] = noise * rnor() + 127.5 + (parityb(sr & V27POLYA) ? Gain : -Gain);
    symbols[sym_p+1] = noise * rnor() + 127.5 + (parityb(sr & V27POLYB) ? Gain : -Gain);
    
    symbols[sym_p] = 0; symbols[sym_p+1] = 0;

    /* Convolutional decode */
    update_viterbi27_blk(vp,&symbols[sym_p],1);
    sym_p = (sym_p + 2) % BUFFERSIZE; /* Must be multiple of 2, be careful when we do r=1/3 */
    
    /* Shift the transmitted data buffer to make room for another bit.
     * There's an extra byte on the end of transmitted_data[] to accomodate the k-1 = 6 bits of delay
     * represented by the encoder state. Each new data bit goes into the 0x04 position of transmitted_data[DECODE_BUF].
     * The oldest transmitted bit is in the 0x80 position of transmitted_data[0], which is lost
     */
    for(i=0;i<DECODE_BUF;i++)
      transmitted_data[i] = (transmitted_data[i] << 1) | (transmitted_data[i+1] >> 7);
    transmitted_data[i] = (transmitted_data[i] << 1) | (databit << 2);

    /* Compare results of chainback from 0 state with chainback from best state */
    chainback_viterbi27(vp,decode_data_0,8*DECODE_BUF,0);
    chainback_viterbi27(vp,decode_data_b,8*DECODE_BUF,-1);

#if 1
    /* Scan both decoder output buffers for bit errors */
    errors = 0;
    for(p=0;p<DECODE_BUF;p++){
      int i;

      errors |= error0 = decode_data_0[p] ^ transmitted_data[p];
      totalerrors_static += popcnt(error0);
      for(i=7; error0 && i >= 0; i--){
	bit_error_counts_0[p*8 + i] += (error0 & 1);
	error0 >>= 1;
      }

      errors |= errorb = decode_data_b[p] ^ transmitted_data[p];
      totalerrors_best += popcnt(errorb);
      for(i=7; errorb && i >= 0; i--){
	bit_error_counts_b[p*8 + i] += (errorb & 1);
	errorb >>= 1;
      }
    }
#endif
#if 0
    if(Verbose){
      errors = 0;
      for(p=0;p<DECODE_BUF;p++)
	errors |= decode_data_0[p] ^ decode_data_b[p];

      if(errors){
	printf("Bit %3d:",bit);
	for(p=0;p<DECODE_BUF;p++){
	  errors = decode_data_0[p] ^ decode_data_b[p];
	  printf(errors ? " %2x" : "   ",errors);
	}
	printf("\n");
      }
    }
#else
    if(Verbose && errors){
      printf("Bit%8d, zero: ",bit);
      errors = 0;
      for(p=0;p<DECODE_BUF;p++){
	errors = decode_data_0[p] ^ transmitted_data[p];
	printf(errors ? " %2x" : "   ",errors);
      }
      putchar('\n');
      printf("             best: ");
      errors = 0;
      for(p=0;p<DECODE_BUF;p++){
	errors = decode_data_b[p] ^ transmitted_data[p];
	printf(errors ? " %2x" : "   ",errors);
      }
      putchar('\n');
    }
#endif
  }
  printf("Total bit errors, chainback from 0 state: %lld\n",totalerrors_static);
  printf("Total bit errors, chainback from best state: %lld\n",totalerrors_best);
  printf("Error counts:\nposition      zero-state           best-state\n");
  for(i=8*DECODE_BUF-1;i >= 0;i--){
    printf("%3d %20llu %20llu\n",i-8*DECODE_BUF,bit_error_counts_0[i],bit_error_counts_b[i]);
  }
  exit(0);
}
